"""
Parse the horrible legacy inventory-related format.

It's typically only used for object contents now.
"""
from __future__ import annotations

import dataclasses
import datetime as dt
import logging
import struct
import typing
import weakref
from io import StringIO
from typing import *

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.legacy_schema import (
    parse_schema_line,
    SchemaBase,
    SchemaDate,
    SchemaFieldSerializer,
    SchemaHexInt,
    SchemaInt,
    SchemaMultilineStr,
    SchemaParsingError,
    SchemaStr,
    SchemaUUID,
    schema_field,
)

MAGIC_ID = UUID("3c115e51-04f4-523c-9fa6-98aff1034730")
LOG = logging.getLogger(__name__)
_T = TypeVar("_T")


class SchemaFlagField(SchemaHexInt):
    """Like a hex int, but must be serialized as bytes in LLSD due to being a U32"""
    @classmethod
    def from_llsd(cls, val: Any) -> int:
        return struct.unpack("!I", val)[0]

    @classmethod
    def to_llsd(cls, val: int) -> Any:
        return struct.pack("!I", val)


def _yield_schema_tokens(reader: StringIO):
    in_bracket = False
    # empty str == EOF in Python
    while line := reader.readline():
        line = line.strip()
        # Whitespace-only lines are automatically skipped
        if not line:
            continue
        try:
            key, val = parse_schema_line(line)
        except SchemaParsingError:
            # Can happen if there's a malformed multi-line string, just
            # skip by it.
            LOG.warning(f"Found invalid inventory line {line!r}")
            continue
        if key == "{":
            if in_bracket:
                LOG.warning("Found multiple opening brackets inside structure, "
                            "was a nested structure not handled?")
            in_bracket = True
            continue
        if key == "}":
            if not in_bracket:
                LOG.warning("Unexpected closing bracket")
            in_bracket = False
            break
        yield key, val
    if in_bracket:
        LOG.warning("Reached EOF while inside a bracket")


class InventoryBase(SchemaBase):
    SCHEMA_NAME: ClassVar[str]

    @classmethod
    def from_reader(cls, reader: StringIO, read_header=False) -> InventoryBase:
        tok_iter = _yield_schema_tokens(reader)
        # Someone else hasn't already read the header for us
        if read_header:
            schema_name, _ = next(tok_iter)
            if schema_name != cls.SCHEMA_NAME:
                raise ValueError(f"Expected schema name {schema_name!r} to be {cls.SCHEMA_NAME!r}")

        fields = cls._get_fields_dict()
        obj_dict = {}
        for key, val in tok_iter:
            if key in fields:
                field: dataclasses.Field = fields[key]
                spec = field.metadata.get("spec")
                # Not a real key, an internal var on our dataclass
                if not spec:
                    LOG.warning(f"Internal key {key!r}")
                    continue
                # some kind of nested structure like sale_info
                if issubclass(spec, SchemaBase):
                    obj_dict[key] = spec.from_reader(reader)
                elif issubclass(spec, SchemaFieldSerializer):
                    obj_dict[key] = spec.deserialize(val)
                else:
                    raise ValueError(f"Unsupported spec for {key!r}, {spec!r}")
            else:
                LOG.warning(f"Unknown key {key!r}")
        return cls._obj_from_dict(obj_dict)

    def to_writer(self, writer: StringIO):
        writer.write(f"\t{self.SCHEMA_NAME}\t0\n")
        writer.write("\t{\n")
        for field_name, field in self._get_fields_dict().items():
            spec = field.metadata.get("spec")
            # Not meant to be serialized
            if not spec:
                continue
            if field.metadata.get("llsd_only"):
                continue

            val = getattr(self, field_name)
            if val is None:
                continue

            # Some kind of nested structure like sale_info
            if isinstance(val, SchemaBase):
                val.to_writer(writer)
            elif issubclass(spec, SchemaFieldSerializer):
                writer.write(f"\t\t{field_name}\t{spec.serialize(val)}\n")
            else:
                raise ValueError(f"Bad inventory spec {spec!r}")
        writer.write("\t}\n")


class InventoryDifferences(typing.NamedTuple):
    changed: List[InventoryNodeBase]
    removed: List[InventoryNodeBase]


class InventoryModel(InventoryBase):
    def __init__(self):
        self.nodes: Dict[UUID, InventoryNodeBase] = {}
        self.root: Optional[InventoryContainerBase] = None

    @classmethod
    def from_reader(cls, reader: StringIO, read_header=False) -> InventoryModel:
        model = cls()
        for key, value in _yield_schema_tokens(reader):
            if key == "inv_object":
                obj = InventoryObject.from_reader(reader)
                if obj is not None:
                    model.add(obj)
            elif key == "inv_category":
                cat = InventoryCategory.from_reader(reader)
                if cat is not None:
                    model.add(cat)
            elif key == "inv_item":
                item = InventoryItem.from_reader(reader)
                if item is not None:
                    model.add(item)
            else:
                LOG.warning("Unknown key {0}".format(key))
        return model

    @classmethod
    def from_llsd(cls, llsd_val: List[Dict]) -> InventoryModel:
        model = cls()
        for obj_dict in llsd_val:
            for inv_type in INVENTORY_TYPES:
                if inv_type.ID_ATTR in obj_dict:
                    if (obj := inv_type.from_llsd(obj_dict)) is not None:
                        model.add(obj)
                    break
                LOG.warning(f"Unknown object type {obj_dict!r}")
        return model

    @property
    def ordered_nodes(self) -> Iterable[InventoryNodeBase]:
        yield from self.all_containers
        yield from self.all_items

    @property
    def all_containers(self) -> Iterable[InventoryContainerBase]:
        for node in self.nodes.values():
            if isinstance(node, InventoryContainerBase):
                yield node

    @property
    def all_items(self) -> Iterable[InventoryItem]:
        for node in self.nodes.values():
            if not isinstance(node, InventoryContainerBase):
                yield node

    def __eq__(self, other):
        if not isinstance(other, InventoryModel):
            return False
        return set(self.nodes.values()) == set(other.nodes.values())

    def to_writer(self, writer: StringIO):
        for node in self.ordered_nodes:
            node.to_writer(writer)

    def to_llsd(self):
        return list(node.to_llsd() for node in self.ordered_nodes)

    def add(self, node: InventoryNodeBase):
        if node.node_id in self.nodes:
            raise KeyError(f"{node.node_id} already exists in the inventory model")

        self.nodes[node.node_id] = node
        if isinstance(node, InventoryContainerBase):
            if node.parent_id == UUID.ZERO:
                self.root = node
        node.model = weakref.proxy(self)

    def unlink(self, node: InventoryNodeBase, single_only: bool = False) -> Sequence[InventoryNodeBase]:
        """Unlink a node and its descendants from the tree, returning the removed nodes"""
        assert node.model == self
        if node == self.root:
            self.root = None
        unlinked = [node]
        if isinstance(node, InventoryContainerBase) and not single_only:
            for child in node.children:
                unlinked.extend(self.unlink(child))
        self.nodes.pop(node.node_id, None)
        node.model = None
        return unlinked

    def get_differences(self, other: InventoryModel) -> InventoryDifferences:
        # Includes modified things with the same ID
        changed_in_other = []
        removed_in_other = []

        other_keys = set(other.nodes.keys())
        our_keys = set(self.nodes.keys())

        # Removed
        for key in our_keys - other_keys:
            removed_in_other.append(self.nodes[key])

        # Updated
        for key in other_keys.intersection(our_keys):
            other_node = other.nodes[key]
            if other_node != self.nodes[key]:
                changed_in_other.append(other_node)

        # Added
        for key in other_keys - our_keys:
            changed_in_other.append(other.nodes[key])
        return InventoryDifferences(
            changed=changed_in_other,
            removed=removed_in_other,
        )

    def __getitem__(self, item: UUID) -> InventoryNodeBase:
        return self.nodes[item]

    def __contains__(self, item: UUID):
        return item in self.nodes

    def get(self, item: UUID) -> Optional[InventoryNodeBase]:
        return self.nodes.get(item)


@dataclasses.dataclass
class InventoryPermissions(InventoryBase):
    SCHEMA_NAME: ClassVar[str] = "permissions"

    base_mask: int = schema_field(SchemaHexInt)
    owner_mask: int = schema_field(SchemaHexInt)
    group_mask: int = schema_field(SchemaHexInt)
    everyone_mask: int = schema_field(SchemaHexInt)
    next_owner_mask: int = schema_field(SchemaHexInt)
    creator_id: UUID = schema_field(SchemaUUID)
    owner_id: UUID = schema_field(SchemaUUID)
    last_owner_id: UUID = schema_field(SchemaUUID)
    group_id: UUID = schema_field(SchemaUUID)
    # Nothing actually cares about this, but it could be there.
    # It's kind of redundant since it just means owner_id == NULL_KEY && group_id != NULL_KEY.
    is_owner_group: int = schema_field(SchemaInt, default=0, llsd_only=True)


@dataclasses.dataclass
class InventorySaleInfo(InventoryBase):
    SCHEMA_NAME: ClassVar[str] = "sale_info"

    sale_type: str = schema_field(SchemaStr)
    sale_price: int = schema_field(SchemaInt)


@dataclasses.dataclass
class InventoryNodeBase(InventoryBase):
    ID_ATTR: ClassVar[str]

    name: str

    parent_id: Optional[UUID] = schema_field(SchemaUUID)
    model: Optional[InventoryModel] = dataclasses.field(
        default=None, init=False, hash=False, compare=False, repr=False
    )

    @property
    def node_id(self) -> UUID:
        return getattr(self, self.ID_ATTR)

    @node_id.setter
    def node_id(self, val: UUID):
        setattr(self, self.ID_ATTR, val)

    @property
    def parent(self) -> Optional[InventoryContainerBase]:
        return self.model.nodes.get(self.parent_id)

    def unlink(self) -> Sequence[InventoryNodeBase]:
        return self.model.unlink(self)

    @classmethod
    def _obj_from_dict(cls, obj_dict):
        # Bad entry, ignore
        # TODO: Check on these. might be symlinks or something.
        if obj_dict.get("type") == "-1":
            LOG.warning(f"Skipping bad object with type == -1: {obj_dict!r}")
            return None
        return super()._obj_from_dict(obj_dict)

    def __hash__(self):
        return hash(self.node_id)

    def __iter__(self) -> Iterator[InventoryNodeBase]:
        return iter(())

    def __contains__(self, item) -> bool:
        return item in tuple(self)


@dataclasses.dataclass
class InventoryContainerBase(InventoryNodeBase):
    type: str = schema_field(SchemaStr)
    name: str = schema_field(SchemaMultilineStr)

    @property
    def children(self) -> Sequence[InventoryNodeBase]:
        return tuple(
            x for x in self.model.nodes.values()
            if x.parent_id == self.node_id
        )

    def __getitem__(self, item: Union[int, str]) -> InventoryNodeBase:
        if isinstance(item, int):
            return self.children[item]

        for child in self.children:
            if child.name == item:
                return child
        raise KeyError(f"{item!r} not found in children")

    def __iter__(self) -> Iterator[InventoryNodeBase]:
        return iter(self.children)

    def get_or_create_subcategory(self, name: str) -> InventoryCategory:
        for child in self:
            if child.name == name and isinstance(child, InventoryCategory):
                return child
        child = InventoryCategory(
            name=name,
            cat_id=UUID.random(),
            parent_id=self.node_id,
            type="category",
            pref_type="-1",
            owner_id=getattr(self, 'owner_id', UUID.ZERO),
            version=1,
        )
        self.model.add(child)
        return child

    # So autogenerated __hash__ doesn't kill our inherited one
    __hash__ = InventoryNodeBase.__hash__


@dataclasses.dataclass
class InventoryObject(InventoryContainerBase):
    SCHEMA_NAME: ClassVar[str] = "inv_object"
    ID_ATTR: ClassVar[str] = "obj_id"

    obj_id: UUID = schema_field(SchemaUUID)

    __hash__ = InventoryNodeBase.__hash__


@dataclasses.dataclass
class InventoryCategory(InventoryContainerBase):
    ID_ATTR: ClassVar[str] = "cat_id"
    SCHEMA_NAME: ClassVar[str] = "inv_category"
    VERSION_NONE: ClassVar[int] = -1

    cat_id: UUID = schema_field(SchemaUUID)
    pref_type: str = schema_field(SchemaStr, llsd_name="preferred_type")
    owner_id: UUID = schema_field(SchemaUUID)
    version: int = schema_field(SchemaInt)

    __hash__ = InventoryNodeBase.__hash__


@dataclasses.dataclass
class InventoryItem(InventoryNodeBase):
    SCHEMA_NAME: ClassVar[str] = "inv_item"
    ID_ATTR: ClassVar[str] = "item_id"

    item_id: UUID = schema_field(SchemaUUID)
    type: str = schema_field(SchemaStr)
    inv_type: str = schema_field(SchemaStr)
    flags: int = schema_field(SchemaFlagField)
    name: str = schema_field(SchemaMultilineStr)
    desc: str = schema_field(SchemaMultilineStr)
    creation_date: dt.datetime = schema_field(SchemaDate, llsd_name="created_at")
    permissions: InventoryPermissions = schema_field(InventoryPermissions)
    sale_info: InventorySaleInfo = schema_field(InventorySaleInfo)
    asset_id: Optional[UUID] = schema_field(SchemaUUID, default=None)
    shadow_id: Optional[UUID] = schema_field(SchemaUUID, default=None)

    __hash__ = InventoryNodeBase.__hash__

    @property
    def true_asset_id(self) -> UUID:
        if self.asset_id is not None:
            return self.asset_id
        return self.shadow_id ^ MAGIC_ID


INVENTORY_TYPES: Tuple[Type[InventoryNodeBase], ...] = (InventoryCategory, InventoryObject, InventoryItem)



from __future__ import annotations

import collections
import dataclasses
import datetime
import functools
import logging
import multiprocessing
import weakref
from typing import *
from weakref import ref

from outleap import LEAPClient

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.helpers import proxify
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.client.state import BaseClientSession
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.circuit import ProxiedCircuit
from hippolyzer.lib.proxy.http_asset_repo import HTTPAssetRepo
from hippolyzer.lib.proxy.http_proxy import HTTPFlowContext
from hippolyzer.lib.proxy.caps import is_asset_server_cap_name, CapData, CapType
from hippolyzer.lib.proxy.inventory_manager import ProxyInventoryManager
from hippolyzer.lib.proxy.namecache import ProxyNameCache
from hippolyzer.lib.proxy.object_manager import ProxyWorldObjectManager
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.settings import ProxySettings

if TYPE_CHECKING:
    from hippolyzer.lib.proxy.message_logger import BaseMessageLogger
    from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow


class Session(BaseClientSession):
    def __init__(self, session_id, secure_session_id, agent_id, circuit_code,
                 session_manager: Optional[SessionManager], login_data=None):
        self.login_data = login_data or {}
        self.pending = True
        self.id: UUID = session_id
        self.secure_session_id: UUID = secure_session_id
        self.agent_id: UUID = agent_id
        self.circuit_code = circuit_code
        self.global_caps = {}
        # Bag of arbitrary data addons can use to persist data across addon reloads
        # Each addon name gets its own separate dict within this dict
        self.addon_ctx: Dict[str, Dict[str, Any]] = collections.defaultdict(dict)
        self.session_manager: SessionManager = session_manager or None
        self.selected: SelectionModel = SelectionModel()
        self.regions: List[ProxiedRegion] = []
        self.started_at = datetime.datetime.now()
        self.message_handler: MessageHandler[Message, str] = MessageHandler()
        self.http_message_handler: MessageHandler[HippoHTTPFlow, str] = MessageHandler()
        self.objects = ProxyWorldObjectManager(self, session_manager.settings, session_manager.name_cache)
        self.inventory = ProxyInventoryManager(proxify(self))
        self.leap_client: Optional[LEAPClient] = None
        # Base path of a newview type cache directory for this session
        self.cache_dir: Optional[str] = None
        self._main_region = None

    @property
    def global_addon_ctx(self):
        if not self.session_manager:
            return {}
        return self.session_manager.addon_ctx

    @classmethod
    def from_login_data(cls, login_data, session_manager):
        sess = Session(
            session_id=UUID(login_data["session_id"]),
            secure_session_id=UUID(login_data["secure_session_id"]),
            agent_id=UUID(login_data["agent_id"]),
            circuit_code=int(login_data["circuit_code"]),
            session_manager=session_manager,
            login_data=login_data,
        )
        appearance_service = login_data.get("agent_appearance_service")
        map_image_service = login_data.get("map-server-url")
        if appearance_service:
            sess.global_caps["AppearanceService"] = appearance_service
        if map_image_service:
            sess.global_caps["MapImageService"] = map_image_service
        # Login data also has details about the initial sim
        sess.register_region(
            circuit_addr=(login_data["sim_ip"], login_data["sim_port"]),
            handle=(login_data["region_x"] << 32) | login_data["region_y"],
            seed_url=login_data["seed_capability"],
        )
        return sess

    @property
    def main_region(self) -> Optional[ProxiedRegion]:
        if self._main_region and self._main_region() in self.regions:
            return self._main_region()
        return None

    @main_region.setter
    def main_region(self, val: ProxiedRegion):
        self._main_region = weakref.ref(val)

    def register_region(self, circuit_addr: Optional[Tuple[str, int]] = None,
                        seed_url: Optional[str] = None,
                        handle: Optional[int] = None) -> ProxiedRegion:
        if not any((circuit_addr, seed_url)):
            raise ValueError("One of circuit_addr and seed_url must be defined!")

        for region in self.regions:
            if region.circuit_addr == circuit_addr:
                if seed_url and region.cap_urls.get("Seed") != seed_url:
                    region.update_caps({"Seed": seed_url})
                if handle:
                    region.handle = handle
                return region
            if seed_url and region.cap_urls.get("Seed") == seed_url:
                return region

        if not circuit_addr:
            raise ValueError("Can't create region without circuit addr!")

        logging.info("Registering region for %r" % (circuit_addr,))
        region = ProxiedRegion(circuit_addr, seed_url, self, handle=handle)
        self.regions.append(region)
        AddonManager.handle_region_registered(self, region)
        return region

    def region_by_circuit_addr(self, circuit_addr) -> Optional[ProxiedRegion]:
        for region in self.regions:
            if region.circuit_addr == circuit_addr and region.circuit:
                return region
        return None

    def region_by_handle(self, handle: int) -> Optional[ProxiedRegion]:
        for region in self.regions:
            if region.handle == handle:
                return region
        return None

    def open_circuit(self, near_addr, circuit_addr, transport):
        for region in self.regions:
            if region.circuit_addr == circuit_addr:
                if not region.circuit or not region.circuit.is_alive:
                    logging_hook = None
                    if self.session_manager.message_logger:
                        logging_hook = functools.partial(
                            self.session_manager.message_logger.log_lludp_message,
                            self,
                            region,
                        )
                    region.circuit = ProxiedCircuit(
                        near_addr, circuit_addr, transport, logging_hook=logging_hook)
                    AddonManager.handle_circuit_created(self, region)
                    return True
                if region.circuit and region.circuit.is_alive:
                    # Whatever, already open
                    logging.debug("Tried to re-open circuit for %r" % (circuit_addr,))
                    return True
        return False

    def resolve_cap(self, url: str) -> Optional[CapData]:
        for cap_name, cap_url in self.global_caps.items():
            if url.startswith(cap_url):
                return CapData(
                    cap_name, None, None, cap_url
                )
        for region in self.regions:
            resolved_cap = region.resolve_cap(url)
            if resolved_cap:
                cap_name, base_url, cap_type = resolved_cap
                # GetMesh and friends can't be tied to a specific session or region
                # (at least on agni) unless we go through a proxy wrapper, since every
                # region just points at the global asset CDN.
                if is_asset_server_cap_name(cap_name) and cap_type != CapType.WRAPPER:
                    return CapData(cap_name, None, None, base_url, cap_type)
                return CapData(cap_name, ref(region), ref(self), base_url, cap_type)
        return None

    def transaction_to_assetid(self, transaction_id: UUID):
        return UUID.combine(transaction_id, self.secure_session_id)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.id)


class SessionManager:
    def __init__(self, settings: ProxySettings):
        self.settings: ProxySettings = settings
        self.sessions: List[Session] = []
        self.shutdown_signal = multiprocessing.Event()
        self.flow_context = HTTPFlowContext()
        self.asset_repo = HTTPAssetRepo()
        self.message_logger: Optional[BaseMessageLogger] = None
        self.addon_ctx: Dict[str, Dict[str, Any]] = collections.defaultdict(dict)
        self.name_cache = ProxyNameCache()
        self.pending_leap_clients: List[LEAPClient] = []

    def create_session(self, login_data) -> Session:
        session = Session.from_login_data(login_data, self)
        self.name_cache.create_subscriptions(
            session.message_handler,
            session.http_message_handler,
        )
        self.sessions.append(session)
        # TODO: less crap way of tying a LEAP client to a session
        while self.pending_leap_clients:
            leap_client = self.pending_leap_clients.pop(-1)
            # Client may have gone bad since it connected
            if not leap_client.connected:
                continue
            logging.info("Assigned LEAP client to session")
            session.leap_client = leap_client
            break
        logging.info("Created %r" % session)
        return session

    def claim_session(self, session_id) -> Optional[Session]:
        for session in self.sessions:
            if session.pending and session.id == session_id:
                logging.info("Claimed %r" % session)
                session.pending = False
                return session
        return None

    def close_session(self, session: Session):
        logging.info("Closed %r" % session)
        session.objects.clear()
        if session.leap_client:
            session.leap_client.disconnect()
        self.sessions.remove(session)

    def resolve_cap(self, url: str) -> Optional["CapData"]:
        for session in self.sessions:
            cap_data = session.resolve_cap(url)
            if cap_data:
                return cap_data
        return CapData()

    async def leap_client_connected(self, leap_client: LEAPClient):
        self.pending_leap_clients.append(leap_client)
        AddonManager.handle_leap_client_added(self, leap_client)


@dataclasses.dataclass
class SelectionModel:
    object_local: Optional[int] = None
    object_locals: Sequence[int] = dataclasses.field(default_factory=list)
    object_full: Optional[UUID] = None
    parcel_local: Optional[int] = None
    parcel_full: Optional[UUID] = None
    script_item: Optional[UUID] = None
    task_item: Optional[UUID] = None



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
from __future__ import annotations

import dataclasses
import logging
import struct
from typing import *

import lazy_object_proxy
import recordclass

from hippolyzer.lib.base.datatypes import Vector3, Quaternion, Vector4, UUID, TaggedUnion
from hippolyzer.lib.base.message.message import Block
from hippolyzer.lib.base.namevalue import NameValueCollection
import hippolyzer.lib.base.serialization as se
import hippolyzer.lib.base.templates as tmpls


class Object(recordclass.RecordClass, use_weakref=True):  # type: ignore
    LocalID: Optional[int] = None
    State: Optional[int] = None
    FullID: Optional[UUID] = None
    CRC: Optional[int] = None
    PCode: Optional[tmpls.PCode] = None
    Material: Optional[tmpls.MCode] = None
    ClickAction: Optional[int] = None
    Scale: Optional[Vector3] = None
    ParentID: Optional[int] = None
    # Actually contains a weakref proxy
    Parent: Optional[Object] = None
    UpdateFlags: Optional[int] = None
    PathCurve: Optional[int] = None
    ProfileCurve: Optional[int] = None
    PathBegin: Optional[int] = None
    PathEnd: Optional[int] = None
    PathScaleX: Optional[int] = None
    PathScaleY: Optional[int] = None
    PathShearX: Optional[int] = None
    PathShearY: Optional[int] = None
    PathTwist: Optional[int] = None
    PathTwistBegin: Optional[int] = None
    PathRadiusOffset: Optional[int] = None
    PathTaperX: Optional[int] = None
    PathTaperY: Optional[int] = None
    PathRevolutions: Optional[int] = None
    PathSkew: Optional[int] = None
    ProfileBegin: Optional[int] = None
    ProfileEnd: Optional[int] = None
    ProfileHollow: Optional[int] = None
    TextureEntry: Optional[tmpls.TextureEntryCollection] = None
    TextureAnim: Optional[tmpls.TextureAnim] = None
    NameValue: Optional[Any] = None
    Data: Optional[Any] = None
    Text: Optional[str] = None
    TextColor: Optional[bytes] = None
    MediaURL: Optional[str] = None
    PSBlock: Optional[Dict] = None
    ExtraParams: Optional[Dict[tmpls.ExtraParamType, Any]] = None
    Sound: Optional[UUID] = None
    OwnerID: Optional[UUID] = None
    SoundGain: Optional[float] = None
    SoundFlags: Optional[int] = None
    SoundRadius: Optional[float] = None
    JointType: Optional[int] = None
    JointPivot: Optional[int] = None
    JointAxisOrAnchor: Optional[int] = None
    TreeSpecies: Optional[int] = None
    ScratchPad: Optional[bytes] = None
    ObjectCosts: Optional[Dict] = None
    ChildIDs: Optional[List[int]] = None
    # Same as parent, contains weakref proxies.
    Children: Optional[List[Object]] = None

    FootCollisionPlane: Optional[Vector4] = None
    Position: Optional[Vector3] = None
    Velocity: Optional[Vector3] = None
    Acceleration: Optional[Vector3] = None
    Rotation: Optional[Quaternion] = None
    AngularVelocity: Optional[Vector3] = None

    # from ObjectProperties
    CreatorID: Optional[UUID] = None
    GroupID: Optional[UUID] = None
    CreationDate: Optional[int] = None
    BaseMask: Optional[int] = None
    OwnerMask: Optional[int] = None
    GroupMask: Optional[int] = None
    EveryoneMask: Optional[int] = None
    NextOwnerMask: Optional[int] = None
    OwnershipCost: Optional[int] = None
    # TaxRate
    SaleType: Optional[int] = None
    SalePrice: Optional[int] = None
    AggregatePerms: Optional[int] = None
    AggregatePermTextures: Optional[int] = None
    AggregatePermTexturesOwner: Optional[int] = None
    Category: Optional[int] = None
    InventorySerial: Optional[int] = None
    ItemID: Optional[UUID] = None
    FolderID: Optional[UUID] = None
    FromTaskID: Optional[UUID] = None
    LastOwnerID: Optional[UUID] = None
    Name: Optional[str] = None
    Description: Optional[str] = None
    TouchName: Optional[str] = None
    SitName: Optional[str] = None
    TextureID: Optional[List[UUID]] = None
    RegionHandle: Optional[int] = None

    def __init__(self, **_kwargs):
        """ set up the object attributes """
        self.ExtraParams = self.ExtraParams or {}  # Variable 1
        self.ObjectCosts = self.ObjectCosts or {}
        self.ChildIDs = []
        # Same as parent, contains weakref proxies.
        self.Children: List[Object] = []

    @property
    def GlobalPosition(self) -> Vector3:
        return handle_to_global_pos(self.RegionHandle) + self.RegionPosition

    @property
    def RegionPosition(self) -> Vector3:
        if not self.ParentID:
            return self.Position

        if not self.Parent:
            raise ValueError("Can't calculate an orphan's RegionPosition")

        # TODO: Cache this and dirty cache if ancestor updates pos?
        return self.Parent.RegionPosition + (self.Position.rotated(self.Parent.RegionRotation))

    @property
    def RegionRotation(self) -> Quaternion:
        if not self.ParentID:
            return self.Rotation

        if not self.Parent:
            raise ValueError("Can't calculate an orphan's RegionRotation")

        # TODO: Cache this and dirty cache if ancestor updates rot?
        return self.Rotation * self.Parent.RegionRotation

    @property
    def AncestorsKnown(self) -> bool:
        obj = self
        while obj.ParentID:
            if not obj.Parent:
                return False
            obj = obj.Parent
        return True

    def update_properties(self, properties: Dict[str, Any]) -> Set[str]:
        """ takes a dictionary of attribute:value and makes it so """
        updated_properties = set()
        for key, val in properties.items():
            if hasattr(self, key):
                old_val = getattr(self, key, dataclasses.MISSING)
                # Don't check equality if we're using a lazy proxy,
                # parsing is deferred until we actually use it.
                if any(isinstance(x, lazy_object_proxy.Proxy) for x in (old_val, val)):
                    # TODO: be smarter about this. Can we store the raw bytes and
                    #  compare those if it's an unparsed object?
                    is_updated = old_val is not val
                else:
                    is_updated = old_val != val
                if is_updated:
                    updated_properties.add(key)
                setattr(self, key, val)
        return updated_properties

    def to_dict(self):
        val = recordclass.asdict(self)
        del val["Children"]
        del val["Parent"]
        return val

    @property
    def Ancestors(self) -> List[Object]:
        obj = self
        ancestors = []
        while obj.Parent:
            obj = obj.Parent
            ancestors.append(obj)
        return ancestors

    @property
    def Descendents(self) -> List[Object]:
        new_children = [self]
        descendents = []
        while new_children:
            to_check = new_children[:]
            new_children.clear()
            for obj in to_check:
                for child in obj.Children:
                    new_children.append(child)
                    descendents.append(child)
        return descendents


def handle_to_gridxy(handle: int) -> Tuple[int, int]:
    return (handle >> 32) // 256, (handle & 0xFFffFFff) // 256


def gridxy_to_handle(x: int, y: int):
    return ((x * 256) << 32) | (y * 256)


def handle_to_global_pos(handle: int) -> Vector3:
    return Vector3(handle >> 32, handle & 0xFFffFFff)


def normalize_object_update(block: Block, handle: int):
    object_data = {
        "RegionHandle": handle,
        "FootCollisionPlane": None,
        "SoundFlags": block["Flags"],
        "SoundGain": block["Gain"],
        "SoundRadius": block["Radius"],
        **dict(block.items()),
        "TextureEntry": block.deserialize_var("TextureEntry", make_copy=False),
        "NameValue": block.deserialize_var("NameValue", make_copy=False),
        "TextureAnim": block.deserialize_var("TextureAnim", make_copy=False),
        "ExtraParams": block.deserialize_var("ExtraParams", make_copy=False) or {},
        "PSBlock": block.deserialize_var("PSBlock", make_copy=False).value,
        "UpdateFlags": block.deserialize_var("UpdateFlags", make_copy=False),
        "State": block.deserialize_var("State", make_copy=False),
        **block.deserialize_var("ObjectData", make_copy=False).value,
    }
    object_data["LocalID"] = object_data.pop("ID")
    # Empty == not updated
    if not object_data["TextureEntry"]:
        object_data.pop("TextureEntry")
    # OwnerID is only set in this packet if a sound is playing. Don't allow
    # ObjectUpdates to clobber _real_ OwnerIDs we had from ObjectProperties
    # with a null UUID.
    if object_data["OwnerID"] == UUID():
        del object_data["OwnerID"]
    del object_data["Flags"]
    del object_data["Gain"]
    del object_data["Radius"]
    del object_data["ObjectData"]
    return object_data


def normalize_terse_object_update(block: Block, handle: int):
    object_data = {
        **block.deserialize_var("Data", make_copy=False),
        **dict(block.items()),
        "TextureEntry": block.deserialize_var("TextureEntry", make_copy=False),
        "RegionHandle": handle,
    }
    object_data["LocalID"] = object_data.pop("ID")
    object_data.pop("Data")
    # Empty == not updated
    if object_data["TextureEntry"] is None:
        object_data.pop("TextureEntry")
    return object_data


def normalize_object_update_compressed_data(data: bytes):
    # Shared by ObjectUpdateCompressed and VOCache case
    compressed = FastObjectUpdateCompressedDataDeserializer.read(data)
    # TODO: ObjectUpdateCompressed doesn't provide a default value for unused
    #  fields, whereas ObjectUpdate and friends do (TextColor, etc.)
    #  need some way to normalize ObjectUpdates so they won't appear to have
    #  changed just because an ObjectUpdate got sent with a default value
    # Only used for determining which sections are present
    del compressed["Flags"]

    # Unlike other ObjectUpdate types, a null value in an ObjectUpdateCompressed
    # always means that there is no value, not that the value hasn't changed
    # from the client's view. Use the default value when that happens.
    ps_block = compressed.pop("PSBlockNew", None)
    if ps_block is None:
        ps_block = compressed.pop("PSBlock", None)
    if ps_block is None:
        ps_block = TaggedUnion(0, None)
    compressed.pop("PSBlock", None)
    if compressed["NameValue"] is None:
        compressed["NameValue"] = NameValueCollection()
    if compressed["Text"] is None:
        compressed["Text"] = b""
        compressed["TextColor"] = b""
    if compressed["MediaURL"] is None:
        compressed["MediaURL"] = b""
    if compressed["AngularVelocity"] is None:
        compressed["AngularVelocity"] = Vector3()
    if compressed["SoundFlags"] is None:
        compressed["SoundFlags"] = 0
        compressed["SoundGain"] = 0.0
        compressed["SoundRadius"] = 0.0
        compressed["Sound"] = UUID()
    if compressed["TextureEntry"] is None:
        compressed["TextureEntry"] = tmpls.TextureEntryCollection()

    object_data = {
        "PSBlock": ps_block.value,
        # Parent flag not set means explicitly un-parented
        "ParentID": compressed.pop("ParentID", None) or 0,
        "LocalID": compressed.pop("ID"),
        **compressed,
    }
    # Don't clobber OwnerID in case the object has a proper one from
    # a previous ObjectProperties. OwnerID isn't expected to be populated
    # on ObjectUpdates unless an attached sound is playing.
    if object_data["OwnerID"] == UUID():
        del object_data["OwnerID"]
    return object_data


def normalize_object_update_compressed(block: Block, handle: int):
    compressed = normalize_object_update_compressed_data(block["Data"])
    compressed["UpdateFlags"] = block.deserialize_var("UpdateFlags", make_copy=False)
    compressed["RegionHandle"] = handle
    return compressed


class SimpleStructReader(se.BufferReader):
    def read_struct(self, spec: struct.Struct, peek=False) -> Tuple[Any, ...]:
        val = spec.unpack_from(self._buffer, self._pos)
        if not peek:
            self._pos += spec.size
        return val

    def read_bytes_null_term(self) -> bytes:
        old_offset = self._pos
        while self._buffer[self._pos] != 0:
            self._pos += 1
        val = self._buffer[old_offset:self._pos]
        self._pos += 1
        return val


class FastObjectUpdateCompressedDataDeserializer:
    HEADER_STRUCT = struct.Struct("<16sIBBIBB3f3f3fI16s")
    ANGULAR_VELOCITY_STRUCT = struct.Struct("<3f")
    PARENT_ID_STRUCT = struct.Struct("<I")
    TREE_SPECIES_STRUCT = struct.Struct("<B")
    DATAPACKER_LEN = struct.Struct("<I")
    COLOR_ADAPTER = tmpls.Color4()
    PARTICLES_OLD = se.TypedBytesFixed(86, tmpls.PSBLOCK_TEMPLATE)
    SOUND_STRUCT = struct.Struct("<16sfBf")
    PRIM_PARAMS_STRUCT = struct.Struct("<BBHHBBBBbbbbbBbHHH")
    ATTACHMENT_STATE_ADAPTER = tmpls.AttachmentStateAdapter(None)

    @classmethod
    def read(cls, data: bytes) -> Dict:
        reader = SimpleStructReader("<", data)
        foo = reader.read_struct(cls.HEADER_STRUCT)
        full_id, local_id, pcode, state, crc, material, click_action, \
            scalex, scaley, scalez, posx, posy, posz, rotx, roty, rotz, \
            flags, owner_id = foo
        scale = Vector3(scalex, scaley, scalez)
        full_id = UUID(bytes=full_id)
        pcode = tmpls.PCode(pcode)
        if pcode == tmpls.PCode.AVATAR:
            state = tmpls.AgentState(state)
        elif pcode == tmpls.PCode.PRIMITIVE:
            state = cls.ATTACHMENT_STATE_ADAPTER.decode(state, None)
        pos = Vector3(posx, posy, posz)
        rot = Quaternion(rotx, roty, rotz)
        owner_id = UUID(bytes=owner_id)
        ang_vel = None
        if flags & tmpls.CompressedFlags.ANGULAR_VELOCITY.value:
            ang_vel = Vector3(*reader.read_struct(cls.ANGULAR_VELOCITY_STRUCT))
        parent_id = None
        if flags & tmpls.CompressedFlags.PARENT_ID.value:
            parent_id = reader.read_struct(cls.PARENT_ID_STRUCT)[0]
        tree_species = None
        if flags & tmpls.CompressedFlags.TREE.value:
            tree_species = reader.read_struct(cls.TREE_SPECIES_STRUCT)[0]
        scratchpad = None
        if flags & tmpls.CompressedFlags.SCRATCHPAD.value:
            scratchpad = reader.read_bytes(reader.read_struct(cls.DATAPACKER_LEN)[0])
        text = None
        text_color = None
        if flags & tmpls.CompressedFlags.TEXT.value:
            text = reader.read_bytes_null_term().decode("utf8")
            text_color = cls.COLOR_ADAPTER.decode(reader.read_bytes(4), ctx=None)
        media_url = None
        if flags & tmpls.CompressedFlags.MEDIA_URL.value:
            media_url = reader.read_bytes_null_term().decode("utf8")
        psblock = None
        if flags & tmpls.CompressedFlags.PARTICLES.value:
            psblock = reader.read(cls.PARTICLES_OLD)
        extra_params = reader.read(tmpls.EXTRA_PARAM_COLLECTION)
        sound, sound_gain, sound_flags, sound_radius = None, None, None, None
        if flags & tmpls.CompressedFlags.SOUND.value:
            sound, sound_gain, sound_flags, sound_radius = reader.read_struct(cls.SOUND_STRUCT)
            sound = UUID(bytes=sound)
            sound_flags = tmpls.SoundFlags(sound_flags)
        name_value = None
        if flags & tmpls.CompressedFlags.NAME_VALUES.value:
            name_value = reader.read(tmpls.NAMEVALUES_TERMINATED_TEMPLATE)
        path_curve, profile_curve, path_begin, path_end, path_scale_x, path_scale_y, \
            path_shear_x, path_shear_y, path_twist, path_twist_begin, path_radius_offset, \
            path_taper_x, path_taper_y, path_revolutions, path_skew, profile_begin, \
            profile_end, profile_hollow = reader.read_struct(cls.PRIM_PARAMS_STRUCT)
        texture_entry = reader.read(tmpls.DATA_PACKER_TE_TEMPLATE)
        texture_anim = None
        if flags & tmpls.CompressedFlags.TEXTURE_ANIM.value:
            texture_anim = reader.read(se.TypedByteArray(se.U32, tmpls.TA_TEMPLATE))
        psblock_new = None
        if flags & tmpls.CompressedFlags.PARTICLES_NEW.value:
            psblock_new = reader.read(tmpls.PSBLOCK_TEMPLATE)

        if len(reader):
            logging.warning(f"{len(reader)} bytes left at end of buffer for compressed {data!r}")

        return {
            "FullID": full_id,
            "ID": local_id,
            "PCode": pcode,
            "State": state,
            "CRC": crc,
            "Material": material,
            "ClickAction": click_action,
            "Scale": scale,
            "Position": pos,
            "Rotation": rot,
            "Flags": flags,
            "OwnerID": owner_id,
            "AngularVelocity": ang_vel,
            "ParentID": parent_id,
            "TreeSpecies": tree_species,
            "ScratchPad": scratchpad,
            "Text": text,
            "TextColor": text_color,
            "MediaURL": media_url,
            "PSBlock": psblock,
            "ExtraParams": extra_params,
            "Sound": sound,
            "SoundGain": sound_gain,
            "SoundFlags": sound_flags,
            "SoundRadius": sound_radius,
            "NameValue": name_value,
            "PathCurve": path_curve,
            "ProfileCurve": profile_curve,
            "PathBegin": path_begin,  # 0 to 1, quanta = 0.01
            "PathEnd": path_end,  # 0 to 1, quanta = 0.01
            "PathScaleX": path_scale_x,  # 0 to 1, quanta = 0.01
            "PathScaleY": path_scale_y,  # 0 to 1, quanta = 0.01
            "PathShearX": path_shear_x,  # -.5 to .5, quanta = 0.01
            "PathShearY": path_shear_y,  # -.5 to .5, quanta = 0.01
            "PathTwist": path_twist,  # -1 to 1, quanta = 0.01
            "PathTwistBegin": path_twist_begin,  # -1 to 1, quanta = 0.01
            "PathRadiusOffset": path_radius_offset,  # -1 to 1, quanta = 0.01
            "PathTaperX": path_taper_x,  # -1 to 1, quanta = 0.01
            "PathTaperY": path_taper_y,  # -1 to 1, quanta = 0.01
            "PathRevolutions": path_revolutions,  # 0 to 3, quanta = 0.015
            "PathSkew": path_skew,  # -1 to 1, quanta = 0.01
            "ProfileBegin": profile_begin,  # 0 to 1, quanta = 0.01
            "ProfileEnd": profile_end,  # 0 to 1, quanta = 0.01
            "ProfileHollow": profile_hollow,  # 0 to 1, quanta = 0.01
            "TextureEntry": texture_entry,
            "TextureAnim": texture_anim,
            "PSBlockNew": psblock_new,
        }



import socket
import struct

from hippolyzer.lib.base.network.transport import SocketUDPTransport, UDPPacket


class SOCKS5UDPTransport(SocketUDPTransport):
    HEADER_STRUCT = struct.Struct("!HBB4sH")

    @classmethod
    def serialize(cls, packet: UDPPacket, force_socks_header: bool = False) -> bytes:
        # Decide whether we need a header based on packet direction
        if packet.outgoing and not force_socks_header:
            return packet.data
        header = cls.HEADER_STRUCT.pack(
            0, 0, 1, socket.inet_aton(packet.far_addr[0]), packet.far_addr[1])
        return header + packet.data

    def send_packet(self, packet: UDPPacket) -> None:
        self.transport.sendto(self.serialize(packet), packet.dst_addr)



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import struct
import typing

from .msgtypes import MsgFrequency
from .data import msg_tmpl
from .template import MessageTemplate
from .template_parser import MessageTemplateParser


DEFAULT_PARSER = MessageTemplateParser(msg_tmpl)


class TemplateDictionary:
    """the dictionary with all known templates"""

    def __init__(self, template_list=None, message_template=None):
        if template_list is None:
            if message_template is None:
                parser = DEFAULT_PARSER
            else:
                parser = MessageTemplateParser(message_template)
            template_list = parser.message_templates

        self.template_list: typing.List[MessageTemplate] = []
        # maps name to template
        self.message_templates = {}

        # maps (freq,num) to template
        self.message_dict = {}

        self.load_templates(template_list)

    def load_templates(self, template_list):
        self.template_list.clear()
        self.template_list.extend(template_list)
        self.message_templates.clear()
        self.message_dict.clear()

        self.build_dictionaries(template_list)
        self.build_message_ids()

    def get_template_list(self):
        return [t.name for t in self.template_list]

    def build_dictionaries(self, template_list):
        for template in template_list:
            self.message_templates[template.name] = template

            # do a mapping of type to a string for easier reference
            frequency_str = ''
            if template.frequency == MsgFrequency.FIXED_FREQUENCY_MESSAGE:
                frequency_str = "Fixed"
            elif template.frequency == MsgFrequency.LOW_FREQUENCY_MESSAGE:
                frequency_str = "Low"
            elif template.frequency == MsgFrequency.MEDIUM_FREQUENCY_MESSAGE:
                frequency_str = "Medium"
            elif template.frequency == MsgFrequency.HIGH_FREQUENCY_MESSAGE:
                frequency_str = "High"

            self.message_dict[(frequency_str,
                               template.msg_num)] = template

    def build_message_ids(self):
        for template in list(self.message_templates.values()):
            frequency = template.frequency
            num_bytes = None
            if frequency == MsgFrequency.FIXED_FREQUENCY_MESSAGE:
                # have to do this because Fixed messages are stored as a long in the template
                num_bytes = b'\xff\xff\xff' + struct.pack("B", template.msg_num)
            elif frequency == MsgFrequency.LOW_FREQUENCY_MESSAGE:
                num_bytes = b'\xff\xff' + struct.pack("!H", template.msg_num)
            elif frequency == MsgFrequency.MEDIUM_FREQUENCY_MESSAGE:
                num_bytes = b'\xff' + struct.pack("B", template.msg_num)
            elif frequency == MsgFrequency.HIGH_FREQUENCY_MESSAGE:
                num_bytes = struct.pack("B", template.msg_num)
            template.msg_freq_num_bytes = num_bytes

    def get_template_by_name(self, template_name) -> typing.Optional[MessageTemplate]:
        return self.message_templates.get(template_name)

    def get_template_by_pair(self, frequency, num) -> typing.Optional[MessageTemplate]:
        return self.message_dict.get((frequency, num))

    def __getitem__(self, name):
        return self.get_template_by_name(name)

    def __contains__(self, item):
        return item in self.message_templates

    def __iter__(self):
        return iter(self.template_list)


DEFAULT_TEMPLATE_DICT = TemplateDictionary()



from __future__ import annotations

import asyncio
import logging
from typing import *

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.templates import PCode
from hippolyzer.lib.client.namecache import NameCache
from hippolyzer.lib.client.object_manager import (
    ClientObjectManager,
    ObjectUpdateType, ClientWorldObjectManager,
)

from hippolyzer.lib.base.objects import Object
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.proxy.settings import ProxySettings
from hippolyzer.lib.proxy.vocache import RegionViewerObjectCacheChain

if TYPE_CHECKING:
    from hippolyzer.lib.proxy.region import ProxiedRegion
    from hippolyzer.lib.proxy.sessions import Session

LOG = logging.getLogger(__name__)


class ProxyObjectManager(ClientObjectManager):
    """
    Object manager for a specific region
    """
    _region: ProxiedRegion

    def __init__(
            self,
            region: ProxiedRegion,
            may_use_vo_cache: bool = False
    ):
        super().__init__(region)
        self.may_use_vo_cache = may_use_vo_cache
        self.cache_loaded = False
        self.object_cache = RegionViewerObjectCacheChain([])
        self._cache_miss_timer: Optional[asyncio.TimerHandle] = None
        self.queued_cache_misses: Set[int] = set()
        region.message_handler.subscribe(
            "RequestMultipleObjects",
            self._handle_request_multiple_objects,
        )

    def load_cache(self):
        if not self.may_use_vo_cache or self.cache_loaded:
            return
        handle = self._region.handle
        if not handle:
            LOG.warning(f"Tried to load cache for {self._region} without a handle")
            return
        self.cache_loaded = True
        self.object_cache = RegionViewerObjectCacheChain.for_region(
            handle=handle,
            cache_id=self._region.cache_id,
            cache_dir=self._region.session().cache_dir,
        )

    def request_missed_cached_objects_soon(self, report_only=False):
        if self._cache_miss_timer:
            self._cache_miss_timer.cancel()
        # Basically debounce. Will only trigger 0.2 seconds after the last time it's invoked to
        # deal with the initial flood of ObjectUpdateCached and the natural lag time between that
        # and the viewers' RequestMultipleObjects messages
        loop = asyncio.get_event_loop_policy().get_event_loop()
        self._cache_miss_timer = loop.call_later(0.2, self._request_missed_cached_objects, report_only)

    def _request_missed_cached_objects(self, report_only: bool):
        self._cache_miss_timer = None
        if not self.queued_cache_misses:
            # All the queued cache misses ended up being satisfied without us
            # having to request them, no need to fire off a request.
            return
        if report_only:
            print(f"Would have automatically requested {self.queued_cache_misses!r}")
        else:
            self.request_objects(self.queued_cache_misses)
        self.queued_cache_misses.clear()

    def clear(self):
        super().clear()
        self.object_cache = RegionViewerObjectCacheChain([])
        self.cache_loaded = False
        self.queued_cache_misses.clear()
        if self._cache_miss_timer:
            self._cache_miss_timer.cancel()
        self._cache_miss_timer = None

    def _is_localid_selected(self, localid: int):
        return localid in self._region.session().selected.object_locals

    def _handle_request_multiple_objects(self, msg: Message):
        # Remove any queued cache misses that the viewer just requested for itself
        self.queued_cache_misses -= {b["ID"] for b in msg["ObjectData"]}


class ProxyWorldObjectManager(ClientWorldObjectManager):
    _session: Session
    _settings: ProxySettings

    def __init__(self, session: Session, settings: ProxySettings, name_cache: Optional[NameCache]):
        super().__init__(session, settings, name_cache)
        session.http_message_handler.subscribe(
            "GetObjectCost",
            self._handle_get_object_cost
        )
        session.http_message_handler.subscribe(
            "FirestormBridge",
            self._handle_firestorm_bridge_request,
        )

    def _handle_object_update_cached_misses(self, region_handle: int, missing_locals: Set[int]):
        region_mgr: Optional[ProxyObjectManager] = self._get_region_manager(region_handle)
        if not self._settings.ALLOW_AUTO_REQUEST_OBJECTS:
            if self._settings.USE_VIEWER_OBJECT_CACHE:
                region_mgr.queued_cache_misses |= missing_locals
                region_mgr.request_missed_cached_objects_soon(report_only=True)
        elif self._settings.AUTOMATICALLY_REQUEST_MISSING_OBJECTS:
            # Schedule these local IDs to be requested soon if the viewer doesn't request
            # them itself. Ideally we could just mutate the CRC of the ObjectUpdateCached
            # to force a CRC cache miss in the viewer, but that appears to cause the viewer
            # to drop the resulting ObjectUpdateCompressed when the CRC doesn't match?
            # It was causing all objects to go missing even though the ObjectUpdateCompressed
            # was received.
            region_mgr: Optional[ProxyObjectManager] = self._get_region_manager(region_handle)
            region_mgr.queued_cache_misses |= missing_locals
            region_mgr.request_missed_cached_objects_soon()

    def _run_object_update_hooks(self, obj: Object, updated_props: Set[str], update_type: ObjectUpdateType,
                                 msg: Optional[Message]):
        super()._run_object_update_hooks(obj, updated_props, update_type, msg)
        region = self._session.region_by_handle(obj.RegionHandle)
        if self._settings.ALLOW_AUTO_REQUEST_OBJECTS:
            if obj.PCode == PCode.AVATAR and "ParentID" in updated_props:
                if obj.ParentID and not region.objects.lookup_localid(obj.ParentID):
                    # If an avatar just sat on an object we don't know about, add it to the queued
                    # cache misses and request it if the viewer doesn't. This should happen
                    # regardless of the auto-request missing objects setting because otherwise we
                    # have no way to get a sitting agent's true region location, even if it's ourselves.
                    region.objects.queued_cache_misses.add(obj.ParentID)
                    region.objects.request_missed_cached_objects_soon()
        AddonManager.handle_object_updated(self._session, region, obj, updated_props, msg)

    def _run_kill_object_hooks(self, obj: Object):
        super()._run_kill_object_hooks(obj)
        region = self._session.region_by_handle(obj.RegionHandle)
        AddonManager.handle_object_killed(self._session, region, obj)

    def _lookup_cache_entry(self, region_handle: int, local_id: int, crc: int) -> Optional[bytes]:
        region_mgr: Optional[ProxyObjectManager] = self._get_region_manager(region_handle)
        return region_mgr.object_cache.lookup_object_data(local_id, crc)

    def _handle_get_object_cost(self, flow: HippoHTTPFlow):
        parsed = llsd.parse_xml(flow.response.content)
        self._process_get_object_cost_response(parsed)

    def _handle_firestorm_bridge_request(self, flow: HippoHTTPFlow):
        """
        Pull guessed avatar Z offsets from Firestorm Bridge requests

        CoarseLocationUpdate packets can only represent heights up to 1024, so
        viewers typically use an LSL bridge to get avatar heights beyond that range
        and combine it with their X and Y coords from CoarseLocationUpdate packets.
        """
        if not flow.request.content.startswith(b'<llsd><string>getZOffsets|'):
            return
        parsed: str = llsd.parse_xml(flow.response.content)
        if not parsed:
            return

        # av_1_id, 1025.001, av_2_id, 3000.0, ...
        split = parsed.split(", ")
        for av_id, z_offset in zip(split[0::2], split[1::2]):
            av_id = UUID(av_id)
            z_offset = float(z_offset)
            av = self.lookup_avatar(av_id)
            if not av:
                continue
            av.GuessedZ = z_offset



import enum
import logging
import typing

from PySide6 import QtCore, QtGui

from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.message_logger import FilteringMessageLogger

LOG = logging.getLogger(__name__)


class MessageLogHeader(enum.IntEnum):
    Host = 0
    Type = enum.auto()
    Method = enum.auto()
    Name = enum.auto()
    Summary = enum.auto()


class MessageLogModel(QtCore.QAbstractTableModel, FilteringMessageLogger):
    def __init__(self, parent=None, maxlen=2000):
        QtCore.QAbstractTableModel.__init__(self, parent)
        FilteringMessageLogger.__init__(self, maxlen=maxlen)

    def _begin_insert(self, insert_idx: int):
        self.beginInsertRows(QtCore.QModelIndex(), insert_idx, insert_idx)

    def _end_insert(self):
        self.endInsertRows()

    def _begin_reset(self):
        self.beginResetModel()

    def _end_reset(self):
        self.endResetModel()

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self._filtered_entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(MessageLogHeader)

    def data(self, index, role=None):
        if not index.isValid():
            return None
        entry = self._filtered_entries[index.row()]
        if role == QtCore.Qt.UserRole:
            return entry
        if role != QtCore.Qt.DisplayRole:
            return None

        col = index.column()
        val = None

        if col == MessageLogHeader.Host:
            val = entry.host
        elif col == MessageLogHeader.Method:
            val = entry.method
        elif col == MessageLogHeader.Type:
            val = entry.type
        elif col == MessageLogHeader.Name:
            val = entry.name
        elif col == MessageLogHeader.Summary:
            val = entry.summary
        return val

    def headerData(self, col, orientation, role=None):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return MessageLogHeader(col).name


class RegionListModel(QtCore.QAbstractListModel):
    def __init__(self, parent, session_manager):
        super().__init__(parent)
        self.sessionManager = session_manager
        session_manager.regionAdded.connect(self.regionAdded)
        session_manager.regionRemoved.connect(self.regionRemoved)
        self.regions: typing.List[ProxiedRegion] = session_manager.all_regions

    @QtCore.Slot(ProxiedRegion)  # noqa
    def regionAdded(self, region: ProxiedRegion):
        if region in self.regions:
            return

        num_regions = len(self.regions)
        self.beginInsertRows(QtCore.QModelIndex(), num_regions, num_regions)
        self.regions.append(region)
        self.endInsertRows()

    @QtCore.Slot(ProxiedRegion)  # noqa
    def regionRemoved(self, region: ProxiedRegion):
        try:
            region_idx = self.regions.index(region)
        except ValueError:
            return

        self.beginRemoveRows(QtCore.QModelIndex(), region_idx, region_idx)
        self.regions.remove(region)
        self.endRemoveRows()

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.regions)

    def columnCount(self, parent: QtCore.QModelIndex) -> int:
        return 1

    def data(self, index, role=None):
        if not index.isValid():
            return None
        cur_region = self.regions[index.row()]
        if role == QtCore.Qt.UserRole:
            return cur_region
        elif role == QtCore.Qt.FontRole:
            for session in self.sessionManager.sessions:
                if cur_region == session.main_region:
                    font = QtGui.QFont()
                    font.setBold(True)
                    # Main region in "default" session
                    if session == self.sessionManager.sessions[-1]:
                        font.setItalic(True)
                    return font
        if role != QtCore.Qt.DisplayRole:
            return None

        col = index.column()
        val = None

        if col == 0:
            if cur_region and cur_region.session():
                val = f"{cur_region.name} ({cur_region.session().agent_id})"
            # Can happen if the weakref becomes invalid
            else:
                val = "<Dead Region>"
        return val

    def headerData(self, col, orientation, role=None):
        return None

    def clear(self):
        self.beginResetModel()
        self.regions = []
        self.endResetModel()



from __future__ import annotations

import abc
import asyncio
import copy
import dataclasses
import datetime as dt
import logging
from typing import *
from typing import Optional

from .message_handler import MessageHandler
from ..network.transport import AbstractUDPTransport, UDPPacket, Direction, ADDR_TUPLE
from .message import Block, Message
from .msgtypes import PacketFlags
from .udpserializer import UDPMessageSerializer


@dataclasses.dataclass
class ReliableResendInfo:
    last_resent: dt.datetime
    message: Message
    completed: asyncio.Future = dataclasses.field(default_factory=asyncio.Future)
    tries_left: int = 10


class Circuit:
    def __init__(self, near_host: Optional[ADDR_TUPLE], far_host: ADDR_TUPLE, transport):
        self.near_host: Optional[ADDR_TUPLE] = near_host
        self.host: ADDR_TUPLE = far_host
        self.is_alive = True
        self.transport: Optional[AbstractUDPTransport] = transport
        self.serializer = UDPMessageSerializer()
        self.last_packet_at = dt.datetime.now()
        self.packet_id_base = 0
        self.unacked_reliable: Dict[Tuple[Direction, int], ReliableResendInfo] = {}
        self.resend_every: float = 3.0

    def _send_prepared_message(self, message: Message, transport=None):
        try:
            serialized = self.serializer.serialize(message)
        except:
            logging.exception(f"Failed to serialize: {message.to_dict()!r}")
            raise
        return self.send_datagram(serialized, message.direction, transport=transport)

    def send_datagram(self, data: bytes, direction: Direction, transport=None):
        self.last_packet_at = dt.datetime.now()
        src_addr, dst_addr = self.host, self.near_host
        if direction == Direction.OUT:
            src_addr, dst_addr = self.near_host, self.host

        packet = UDPPacket(src_addr, dst_addr, data, direction)
        (transport or self.transport).send_packet(packet)
        return packet

    def prepare_message(self, message: Message):
        if message.finalized:
            raise RuntimeError(f"Trying to re-send finalized {message!r}")
        message.packet_id = self.packet_id_base
        self.packet_id_base += 1
        if message.acks:
            message.send_flags |= PacketFlags.ACK
        else:
            message.send_flags &= ~PacketFlags.ACK
        # If it was queued, it's not anymore
        message.queued = False
        message.finalized = True

    def send(self, message: Message, transport=None) -> UDPPacket:
        if self.prepare_message(message):
            # If the message originates from us then we're responsible for resends.
            if message.reliable and message.synthetic:
                self.unacked_reliable[(message.direction, message.packet_id)] = ReliableResendInfo(
                    last_resent=dt.datetime.now(),
                    message=message,
                )
            return self._send_prepared_message(message, transport)

    def send_reliable(self, message: Message, transport=None) -> asyncio.Future:
        """send() wrapper that always sends reliably and allows `await`ing ACK receipt"""
        if not message.synthetic:
            raise ValueError("Not able to send non-synthetic message reliably!")
        message.send_flags |= PacketFlags.RELIABLE
        self.send(message, transport)
        return self.unacked_reliable[(message.direction, message.packet_id)].completed

    def collect_acks(self, message: Message):
        effective_acks = list(message.acks)
        if message.name == "PacketAck":
            effective_acks.extend(x["ID"] for x in message["Packets"])
        for ack in effective_acks:
            resend_info = self.unacked_reliable.pop((~message.direction, ack), None)
            if resend_info:
                resend_info.completed.set_result(None)

    def resend_unacked(self):
        for resend_info in list(self.unacked_reliable.values()):
            # Not time to attempt a resend yet
            if dt.datetime.now() - resend_info.last_resent < dt.timedelta(seconds=self.resend_every):
                continue

            msg = copy.copy(resend_info.message)
            resend_info.tries_left -= 1
            # We were on our last try and we never received an ack
            if not resend_info.tries_left:
                logging.warning(f"Giving up on unacked {msg.packet_id}")
                del self.unacked_reliable[(msg.direction, msg.packet_id)]
                resend_info.completed.set_exception(TimeoutError("Exceeded resend limit"))
                continue
            resend_info.last_resent = dt.datetime.now()
            msg.send_flags |= PacketFlags.RESENT
            self._send_prepared_message(msg)

    def send_acks(self, to_ack: Sequence[int], direction=Direction.OUT, packet_id=None):
        logging.debug("%r acking %r" % (direction, to_ack))
        # TODO: maybe tack this onto `.acks` for next message?
        message = Message('PacketAck', *[Block('Packets', ID=x) for x in to_ack])
        message.packet_id = packet_id
        message.direction = direction
        self.send(message)

    def __repr__(self):
        return "<%s %r : %r>" % (self.__class__.__name__, self.near_host, self.host)


class ConnectionHolder(abc.ABC):
    """
    Any object that has both a circuit and a message handler

    Preferred to explicitly passing around a circuit, message handler pair
    because generally a ConnectionHolder represents a region or a client.
    The same region or client may have multiple different circuits across the
    lifetime of a session (due to region restarts, etc.)
    """
    circuit: Optional[Circuit]
    message_handler: MessageHandler[Message, str]



"""
Global context for use within Addon callbacks
"""
from __future__ import annotations

import contextlib
from contextvars import ContextVar, Token
from typing import *

if TYPE_CHECKING:
    from hippolyzer.lib.proxy.region import ProxiedRegion
    from hippolyzer.lib.proxy.sessions import Session

# By using ContextVar, coroutines retain the context as it was when
# they were created.
session: ContextVar[Optional[Session]] = ContextVar("session", default=None)
region: ContextVar[Optional[ProxiedRegion]] = ContextVar("region", default=None)


@contextlib.contextmanager
def push(new_session: Optional[Session] = None, new_region: Optional[ProxiedRegion] = None):
    session_token: Token = session.set(new_session)
    region_token: Token = region.set(new_region)
    try:
        yield
    finally:
        session.reset(session_token)
        region.reset(region_token)



from __future__ import annotations

import re
import sys
from typing import *

from hippolyzer.lib.base.network.caps_client import CapsClient, CAPS_DICT
from hippolyzer.lib.proxy.settings import ProxySettings

if TYPE_CHECKING:
    from hippolyzer.lib.proxy.region import ProxiedRegion


class ProxyCapsClient(CapsClient):
    def __init__(self, settings: ProxySettings, region: Optional[ProxiedRegion] = None):
        super().__init__(None)
        self._region = region
        self._settings = settings

    def _get_caps(self) -> Optional[CAPS_DICT]:
        if not self._region:
            return None
        return self._region.cap_urls

    def _request_fixups(self, cap_or_url: str, headers: Dict, proxy: Optional[bool], ssl: Any):
        # We want to proxy this through Hippolyzer
        if proxy is None:
            # We go through the proxy by default, tack on a header letting mitmproxy know the
            # request came from us so we can tag the request as injected. The header will be popped
            # off before passing through to the server.
            if "X-Hippo-Injected" not in headers:
                headers["X-Hippo-Injected"] = "1"
            proxy_port = self._settings.HTTP_PROXY_PORT
            proxy = f"http://127.0.0.1:{proxy_port}"
            # TODO: set up the SSLContext to validate mitmproxy's cert
            ssl = ssl or False
            # https://github.com/aio-libs/aiohttp/issues/4536
            # https://github.com/aio-libs/aiohttp/issues/4268
            # TLS + HTTP proxy is broken under proactor event loops under windows
            # qasync's event loop is based on the proactor event loop under windows.
            # aiohttp has no intention of fixing this anytime soon and I don't
            # want to screw around with qasync, so rewrite the URL to be HTTP
            # and ask mitmproxy to transparently rewrite the request URI to HTTPS
            # Will break on 3xx redirs but you can write a new qasync event loop wrapper
            # if you care about Windows. I don't.
            if sys.platform == "win32" and cap_or_url.startswith("https:"):
                headers["X-Hippo-Windows-SSL-Hack"] = "1"
                cap_or_url = re.sub(r"^https:", "http:", cap_or_url)
        return cap_or_url, headers, proxy, ssl



from typing import NamedTuple, Union, Optional, List

import hippolyzer.lib.base.serialization as se
from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.mesh import MeshAsset, LLMeshSerializer
from hippolyzer.lib.base.templates import AssetType
from hippolyzer.lib.client.state import BaseClientRegion


class UploadError(Exception):
    pass


class UploadToken(NamedTuple):
    linden_cost: int
    uploader_url: str
    payload: bytes


class MeshUploadDetails(NamedTuple):
    mesh_bytes: bytes
    num_faces: int


class AssetUploader:
    def __init__(self, region: BaseClientRegion):
        self._region = region

    async def initiate_asset_upload(self, name: str, asset_type: AssetType,
                                    body: bytes, flags: Optional[int] = None) -> UploadToken:
        payload = {
            "asset_type": asset_type.human_name,
            "description": "(No Description)",
            "everyone_mask": 0,
            "group_mask": 0,
            "folder_id": UUID.ZERO,  # Puts it in the default folder, I guess. Undocumented.
            "inventory_type": asset_type.inventory_type.human_name,
            "name": name,
            "next_owner_mask": 581632,
        }
        if flags is not None:
            payload['flags'] = flags
        resp_payload = await self._make_newfileagentinventory_req(payload)

        return UploadToken(resp_payload["upload_price"], resp_payload["uploader"], body)

    async def _make_newfileagentinventory_req(self, payload: dict):
        async with self._region.caps_client.post("NewFileAgentInventory", llsd=payload) as resp:
            resp.raise_for_status()
            resp_payload = await resp.read_llsd()
        # Need to sniff the resp payload for this because SL sends a 200 status code on error
        if "error" in resp_payload:
            raise UploadError(resp_payload)
        return resp_payload

    async def complete_upload(self, token: UploadToken) -> dict:
        async with self._region.caps_client.post(token.uploader_url, data=token.payload) as resp:
            resp.raise_for_status()
            resp_payload = await resp.read_llsd()
        # The actual upload endpoints return 200 on error, have to sniff the payload to figure
        # out if it actually failed...
        if "error" in resp_payload:
            raise UploadError(resp_payload)
        await self._handle_upload_complete(resp_payload)
        return resp_payload

    async def _handle_upload_complete(self, resp_payload: dict):
        """
        Generic hook called when any asset upload completes.

        Could trigger an AIS fetch to send the viewer details about the item we just created,
        assuming we were in proxy context.
        """
        pass

    # The mesh upload flow is a little special, so it gets its own method
    async def initiate_mesh_upload(self, name: str, mesh: Union[MeshUploadDetails, MeshAsset],
                                   flags: Optional[int] = None) -> UploadToken:
        if isinstance(mesh, MeshAsset):
            writer = se.BufferWriter("!")
            writer.write(LLMeshSerializer(), mesh)
            mesh = MeshUploadDetails(writer.copy_buffer(), len(mesh.segments['high_lod']))

        asset_resources = self._build_asset_resources(name, [mesh])
        payload = {
            'asset_resources': asset_resources,
            'asset_type': 'mesh',
            'description': '(No Description)',
            'everyone_mask': 0,
            'folder_id': UUID.ZERO,
            'group_mask': 0,
            'inventory_type': 'object',
            'name': name,
            'next_owner_mask': 581632,
            'texture_folder_id': UUID.ZERO
        }
        if flags is not None:
            payload['flags'] = flags
        resp_payload = await self._make_newfileagentinventory_req(payload)

        upload_body = llsd.format_xml(asset_resources)
        return UploadToken(resp_payload["upload_price"], resp_payload["uploader"], upload_body)

    def _build_asset_resources(self, name: str, meshes: List[MeshUploadDetails]) -> dict:
        instances = []
        for mesh in meshes:
            instances.append({
                'face_list': [{
                    'diffuse_color': [1.0, 1.0, 1.0, 1.0],
                    'fullbright': False
                }] * mesh.num_faces,
                'material': 3,
                'mesh': 0,
                'mesh_name': name,
                'physics_shape_type': 2,
                'position': [0.0, 0.0, 0.0],
                'rotation': [0.7071067690849304, 0.0, 0.0, 0.7071067690849304],
                'scale': [1.0, 1.0, 1.0]
            })

        return {
            'instance_list': instances,
            'mesh_list': [mesh.mesh_bytes for mesh in meshes],
            'metric': 'MUT_Unspecified',
            'texture_list': []
        }



# Most things have moved to base's `templates.py` now. This exists
# as an alias module to ease transition.

import dataclasses
import importlib
import logging
from typing import *  # noqa

import hippolyzer.lib.base.templates

try:
    # Most of our templates come from here, so we need to
    # reload this to pull in fresh templates with `import *`
    importlib.reload(hippolyzer.lib.base.templates)  # type: ignore
except:
    logging.exception("Failed to reload base templates lib")


from hippolyzer.lib.base.templates import *


@dataclasses.dataclass
class CAPTemplate:
    cap_name: str
    method: str
    body: Any
    query: Set[str] = dataclasses.field(default_factory=set)
    path: str = ""


# Cap request templates for message builder prefills
CAP_TEMPLATES: List[CAPTemplate] = [
    CAPTemplate(cap_name='Seed', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<array>\n    <string></string>\n  </array>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='SimulatorFeatures', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='EventQueueGet', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>ack</key>\n    <undef />\n    <key>done</key>\n    <boolean></boolean>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='NavMeshGenerationStatus', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='AgentState', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='AgentPreferences', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>default_object_perm_masks</key>\n    <map>\n      <key>Everyone</key>\n      <integer></integer>\n      <key>Group</key>\n      <integer></integer>\n      <key>NextOwner</key>\n      <integer></integer>\n    </map>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='ExtEnvironment', method='GET', body=b'', query={'parcelid'}),
    CAPTemplate(cap_name='UpdateAgentLanguage', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>language</key>\n    <string>en</string>\n    <key>language_is_public</key>\n    <integer>1</integer>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='DirectDelivery', method='GET', body=b'', query=set(), path="/listings"),
    CAPTemplate(cap_name='ViewerStats', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>AgentPositionSnaps</key>\n    <real></real>\n    <key>DisplayNamesEnabled</key>\n    <integer></integer>\n    <key>DisplayNamesShowUsername</key>\n    <integer></integer>\n    <key>MinimalSkin</key>\n    <boolean></boolean>\n    <key>agent</key>\n    <map>\n      <key>agents_in_view</key>\n      <integer></integer>\n      <key>fps</key>\n      <real></real>\n      <key>language</key>\n      <string></string>\n      <key>mem_use</key>\n      <real></real>\n      <key>meters_traveled</key>\n      <real></real>\n      <key>ping</key>\n      <real></real>\n      <key>regions_visited</key>\n      <integer></integer>\n      <key>run_time</key>\n      <real></real>\n      <key>sim_fps</key>\n      <real></real>\n      <key>start_time</key>\n      <integer></integer>\n      <key>version</key>\n      <string></string>\n    </map>\n    <key>downloads</key>\n    <map>\n      <key>mesh_kbytes</key>\n      <real></real>\n      <key>object_kbytes</key>\n      <real></real>\n      <key>texture_kbytes</key>\n      <real></real>\n      <key>world_kbytes</key>\n      <real></real>\n    </map>\n    <key>misc</key>\n    <map>\n      <key>Version</key>\n      <integer></integer>\n      <key>Vertex Buffers Enabled</key>\n      <real></real>\n    </map>\n    <key>session_id</key>\n    <uuid></uuid>\n    <key>stats</key>\n    <map>\n      <key>failures</key>\n      <map>\n        <key>dropped</key>\n        <integer></integer>\n        <key>failed_resends</key>\n        <integer></integer>\n        <key>invalid</key>\n        <integer></integer>\n        <key>missing_updater</key>\n        <integer></integer>\n        <key>off_circuit</key>\n        <integer></integer>\n        <key>resent</key>\n        <integer></integer>\n        <key>send_packet</key>\n        <integer></integer>\n      </map>\n      <key>misc</key>\n      <map>\n        <key>int_1</key>\n        <integer></integer>\n        <key>int_2</key>\n        <integer></integer>\n        <key>string_1</key>\n        <string></string>\n        <key>string_2</key>\n        <string></string>\n      </map>\n      <key>net</key>\n      <map>\n        <key>in</key>\n        <map>\n          <key>compressed_packets</key>\n          <integer></integer>\n          <key>kbytes</key>\n          <real></real>\n          <key>packets</key>\n          <integer></integer>\n          <key>savings</key>\n          <real></real>\n        </map>\n        <key>out</key>\n        <map>\n          <key>compressed_packets</key>\n          <integer></integer>\n          <key>kbytes</key>\n          <real></real>\n          <key>packets</key>\n          <integer></integer>\n          <key>savings</key>\n          <real></real>\n        </map>\n      </map>\n      <key>voice</key>\n      <map>\n        <key>connect_attempts</key>\n        <integer></integer>\n        <key>connect_cycles</key>\n        <integer></integer>\n        <key>connect_time</key>\n        <real></real>\n        <key>establish_attempts</key>\n        <integer></integer>\n        <key>establish_time</key>\n        <real></real>\n        <key>provision_attempts</key>\n        <integer></integer>\n        <key>provision_time</key>\n        <real></real>\n      </map>\n    </map>\n    <key>system</key>\n    <map>\n      <key>address_size</key>\n      <integer></integer>\n      <key>cpu</key>\n      <string></string>\n      <key>gl</key>\n      <map>\n        <key>ati_offset_vertical_lines</key>\n        <integer></integer>\n        <key>ati_old_driver</key>\n        <integer></integer>\n        <key>debug_gpu</key>\n        <integer></integer>\n        <key>gl_renderer</key>\n        <string></string>\n        <key>gpu_vendor</key>\n        <string></string>\n        <key>gpu_version</key>\n        <string></string>\n        <key>has_anisotropic</key>\n        <integer></integer>\n        <key>has_arb_env_combine</key>\n        <integer></integer>\n        <key>has_ati_mem_info</key>\n        <integer></integer>\n        <key>has_blend_func_separate</key>\n        <integer></integer>\n        <key>has_compressed_textures</key>\n        <integer></integer>\n        <key>has_cube_map</key>\n        <integer></integer>\n        <key>has_debug_output</key>\n        <integer></integer>\n        <key>has_depth_clamp</key>\n        <integer></integer>\n        <key>has_draw_buffers</key>\n        <integer></integer>\n        <key>has_flush_buffer_range</key>\n        <integer></integer>\n        <key>has_fragment_shader</key>\n        <integer></integer>\n        <key>has_framebuffer_object</key>\n        <integer></integer>\n        <key>has_map_buffer_range</key>\n        <integer></integer>\n        <key>has_mip_map_generation</key>\n        <integer></integer>\n        <key>has_multitexture</key>\n        <integer></integer>\n        <key>has_nvx_mem_info</key>\n        <integer></integer>\n        <key>has_occlusion_query</key>\n        <integer></integer>\n        <key>has_occlusion_query2</key>\n        <integer></integer>\n        <key>has_pbuffer</key>\n        <integer></integer>\n        <key>has_point_parameters</key>\n        <integer></integer>\n        <key>has_requirements</key>\n        <integer></integer>\n        <key>has_separate_specular_color</key>\n        <integer></integer>\n        <key>has_shader_objects</key>\n        <integer></integer>\n        <key>has_srgb_framebuffer</key>\n        <integer></integer>\n        <key>has_srgb_texture</key>\n        <integer></integer>\n        <key>has_sync</key>\n        <integer></integer>\n        <key>has_texture_multisample</key>\n        <integer></integer>\n        <key>has_texture_rectangle</key>\n        <integer></integer>\n        <key>has_texture_srgb_decode</key>\n        <integer></integer>\n        <key>has_timer_query</key>\n        <integer></integer>\n        <key>has_transform_feedback</key>\n        <integer></integer>\n        <key>has_vertex_array_object</key>\n        <integer></integer>\n        <key>has_vertex_buffer_object</key>\n        <integer></integer>\n        <key>has_vertex_shader</key>\n        <integer></integer>\n        <key>is_ati</key>\n        <integer></integer>\n        <key>is_gf2or4mx</key>\n        <integer></integer>\n        <key>is_gf3</key>\n        <integer></integer>\n        <key>is_gf_gfx</key>\n        <integer></integer>\n        <key>is_intel</key>\n        <integer></integer>\n        <key>is_nvidia</key>\n        <integer></integer>\n        <key>max_color_texture_samples</key>\n        <integer></integer>\n        <key>max_depth_texture_samples</key>\n        <integer></integer>\n        <key>max_index_range</key>\n        <integer></integer>\n        <key>max_integer_samples</key>\n        <integer></integer>\n        <key>max_sample_mask_words</key>\n        <integer></integer>\n        <key>max_samples</key>\n        <integer></integer>\n        <key>max_texture_size</key>\n        <integer></integer>\n        <key>max_vertex_range</key>\n        <integer></integer>\n        <key>num_texture_image_units</key>\n        <integer></integer>\n        <key>num_texture_units</key>\n        <integer></integer>\n        <key>opengl_version</key>\n        <string></string>\n        <key>vram</key>\n        <integer></integer>\n      </map>\n      <key>gpu</key>\n      <string></string>\n      <key>gpu_class</key>\n      <integer></integer>\n      <key>gpu_vendor</key>\n      <string></string>\n      <key>gpu_version</key>\n      <string></string>\n      <key>mac_address</key>\n      <string></string>\n      <key>opengl_version</key>\n      <string></string>\n      <key>os</key>\n      <string></string>\n      <key>ram</key>\n      <integer></integer>\n      <key>serial_number</key>\n      <string></string>\n      <key>shader_level</key>\n      <integer></integer>\n    </map>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='GetDisplayNames', method='GET', body=b'', query={'ids'}),
    CAPTemplate(cap_name='FetchInventory2', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>agent_id</key>\n    <uuid><!HIPPOREPL[[AGENT_ID]]></uuid>\n    <key>cap_name</key>\n    <string></string>\n    <key>items</key>\n    <array>\n      <map>\n        <key>item_id</key>\n        <uuid></uuid>\n        <key>owner_id</key>\n        <uuid><!HIPPOREPL[[AGENT_ID]]></uuid>\n      </map>\n    </array>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='FetchLib2', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>agent_id</key>\n    <uuid></uuid>\n    <key>cap_name</key>\n    <string></string>\n    <key>items</key>\n    <array>\n      <map>\n        <key>item_id</key>\n        <uuid></uuid>\n        <key>owner_id</key>\n        <uuid></uuid>\n      </map>\n    </array>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='FetchInventoryDescendents2', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>folders</key>\n    <array>\n      <map>\n        <key>fetch_folders</key>\n        <boolean></boolean>\n        <key>fetch_items</key>\n        <boolean></boolean>\n        <key>folder_id</key>\n        <string></string>\n        <key>owner_id</key>\n        <uuid><!HIPPOREPL[[AGENT_ID]]></uuid>\n        <key>sort_order</key>\n        <integer></integer>\n      </map>\n    </array>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='ProvisionVoiceAccountRequest', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<undef />\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='AvatarRenderInfo', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='ReadOfflineMsgs', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='MapImageService', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='RenderMaterials', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='AvatarRenderInfo', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>agents</key>\n    <map>\n      <key><!HIPPOREPL[[AGENT_ID]]></key>\n      <map>\n        <key>tooComplex</key>\n        <boolean></boolean>\n        <key>weight</key>\n        <integer></integer>\n      </map>\n    </map>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='ParcelVoiceInfoRequest', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<undef />\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='ProductInfoRequest', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='ObjectMedia', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>object_id</key>\n    <uuid><!HIPPOREPL[[SELECTED_FULL]]></uuid>\n    <key>verb</key>\n    <string>GET</string>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='GetObjectCost', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>object_ids</key>\n    <array>\n      <uuid><!HIPPOREPL[[SELECTED_FULL]]></uuid>\n    </array>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='GetObjectPhysicsData', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>object_ids</key>\n    <array>\n      <uuid><!HIPPOREPL[[SELECTED_FULL]]></uuid>\n    </array>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='RenderMaterials', method='PUT', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>Zipped</key>\n    <binary></binary>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='RenderMaterials', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>Zipped</key>\n    <binary></binary>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='GetExperienceInfo', method='GET', body=b'', query={'public_id', 'page_size'}, path="/id/"),
    CAPTemplate(cap_name='LSLSyntax', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='GetCreatorExperiences', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='GetMetadata', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>fields</key>\n    <array>\n      <string>experience</string>\n    </array>\n    <key>item-id</key>\n    <uuid><!HIPPOREPL[[SELECTED_SCRIPT_ITEM]]></uuid>\n    <key>object-id</key>\n    <uuid><!HIPPOREPL[[SELECTED_FULL]]></uuid>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='UpdateScriptTask', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>experience</key>\n    <uuid></uuid>\n    <key>is_script_running</key>\n    <integer>1</integer>\n    <key>item_id</key>\n    <uuid><!HIPPOREPL[[SELECTED_SCRIPT_ITEM]]></uuid>\n    <key>target</key>\n    <string>mono</string>\n    <key>task_id</key>\n    <uuid><!HIPPOREPL[[SELECTED_FULL]]></uuid>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='EstateAccess', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='RegionExperiences', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='SimConsoleAsync', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<string></string>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='FindExperienceByName', method='GET', body=b'', query={'query', 'page', 'page_size'}),
    CAPTemplate(cap_name='ViewerMetrics', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>agent_id</key>\n    <uuid></uuid>\n    <key>break</key>\n    <boolean></boolean>\n    <key>duration</key>\n    <real></real>\n    <key>initial</key>\n    <boolean></boolean>\n    <key>message</key>\n    <string></string>\n    <key>nearby</key>\n    <map>\n      <key>cloud</key>\n      <integer></integer>\n      <key>downloading</key>\n      <integer></integer>\n      <key>full</key>\n      <integer></integer>\n      <key>gray</key>\n      <integer></integer>\n    </map>\n    <key>rez_status</key>\n    <string></string>\n    <key>sequence</key>\n    <integer></integer>\n    <key>session_id</key>\n    <uuid></uuid>\n    <key>timers</key>\n    <array>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n      <map>\n        <key>central_bake_version</key>\n        <integer></integer>\n        <key>completed</key>\n        <boolean></boolean>\n        <key>grid_x</key>\n        <integer></integer>\n        <key>grid_y</key>\n        <integer></integer>\n        <key>is_self</key>\n        <boolean></boolean>\n        <key>is_using_server_bakes</key>\n        <boolean></boolean>\n        <key>stats</key>\n        <map>\n          <key>count</key>\n          <integer></integer>\n          <key>max</key>\n          <real></real>\n          <key>mean</key>\n          <real></real>\n          <key>min</key>\n          <real></real>\n          <key>std_dev</key>\n          <real></real>\n        </map>\n        <key>timer_name</key>\n        <string></string>\n      </map>\n    </array>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='InventoryAPIv3', method='PATCH', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>name</key>\n    <string></string>\n  </map>\n</llsd>\n', query={'tid'}, path="/item/SOME_ID"),
    CAPTemplate(cap_name='RemoteParcelRequest', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>location</key>\n    <array>\n      <real></real>\n      <real></real>\n      <real></real>\n    </array>\n    <key>region_handle</key>\n    <binary></binary>\n    <key>region_id</key>\n    <uuid></uuid>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='ParcelPropertiesUpdate', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n  <map>\n   <key>local_id</key>\n    <integer><!HIPPOREPL[[SELECTED_PARCEL_LOCAL]]></integer>\n   <key>any_av_sounds</key>\n    <boolean>true</boolean>\n   <key>auth_buyer_id</key>\n    <uuid></uuid>\n   <key>auto_scale</key>\n    <integer>0</integer>\n   <key>category</key>\n    <integer>0</integer>\n   <key>description</key>\n    <string></string>\n   <key>flags</key>\n    <binary>AAAAAQ==</binary>\n   <key>group_av_sounds</key>\n    <boolean>true</boolean>\n   <key>group_id</key>\n    <uuid></uuid>\n   <key>landing_type</key>\n    <integer>2</integer>\n   <key>media_allow_navigate</key>\n    <integer>1</integer>\n   <key>media_current_url</key>\n    <string></string>\n   <key>media_desc</key>\n    <string></string>\n   <key>media_height</key>\n    <integer>0</integer>\n   <key>media_id</key>\n    <uuid></uuid>\n   <key>media_loop</key>\n    <integer>0</integer>\n   <key>media_prevent_camera_zoom</key>\n    <integer>0</integer>\n   <key>media_type</key>\n    <string>none/none</string>\n   <key>media_url</key>\n    <string></string>\n   <key>media_url_timeout</key>\n    <real>0.0</real>\n   <key>media_width</key>\n    <integer>0</integer>\n   <key>music_url</key>\n    <string></string>\n   <key>name</key>\n    <string></string>\n   <key>obscure_media</key>\n    <boolean>false</boolean>\n   <key>obscure_music</key>\n    <boolean>false</boolean>\n   <key>parcel_flags</key>\n    <binary>fiQASw==</binary>\n   <key>pass_hours</key>\n    <real>1.0</real>\n   <key>pass_price</key>\n    <integer>10</integer>\n   <key>sale_price</key>\n    <integer>10000</integer>\n   <key>see_avs</key>\n    <boolean>true</boolean>\n   <key>snapshot_id</key>\n    <uuid></uuid>\n   <key>user_location</key>\n    <array>\n      <real>0.0</real>\n      <real>0.0</real>\n      <real>0.0</real>\n    </array>\n   <key>user_look_at</key>\n    <array>\n      <real>0.0</real>\n      <real>0.0</real>\n      <real>0.0</real>\n    </array>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='UpdateScriptAgent', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>item_id</key>\n    <uuid></uuid>\n    <key>target</key>\n    <string>mono</string>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='NewFileAgentInventory', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>asset_type</key>\n    <string></string>\n    <key>description</key>\n    <string></string>\n    <key>everyone_mask</key>\n    <integer></integer>\n    <key>folder_id</key>\n    <uuid></uuid>\n    <key>group_mask</key>\n    <integer></integer>\n    <key>inventory_type</key>\n    <string></string>\n    <key>name</key>\n    <string></string>\n    <key>next_owner_mask</key>\n    <integer></integer>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='UpdateNotecardAgentInventory', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>item_id</key>\n    <uuid></uuid>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='UpdateNotecardTaskInventory', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>item_id</key>\n    <uuid><!HIPPOREPL[[SELECTED_TASK_ITEM]]></uuid>\n    <key>task_id</key>\n    <uuid><!HIPPOREPL[[SELECTED_FULL]]></uuid>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='UpdateGestureAgentInventory', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>item_id</key>\n    <uuid></uuid>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='UpdateGestureTaskInventory', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>item_id</key>\n    <uuid><!HIPPOREPL[[SELECTED_TASK_ITEM]]></uuid>\n    <key>task_id</key>\n    <uuid><!HIPPOREPL[[SELECTED_FULL]]></uuid>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='MeshUploadFlag', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='UntrustedSimulatorMessage', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n   <key>message</key>\n    <string>ChatFromViewer</string>\n   <key>body</key>\n    <map>\n     <key>AgentData</key>\n      <array>\n        <map>\n         <key>AgentID</key>\n          <uuid><!HIPPOREPL[[AGENT_ID]]></uuid>\n         <key>SessionID</key>\n          <uuid><!HIPPOREPL[[SESSION_ID]]></uuid>\n        </map>\n      </array>\n     <key>ChatData</key>\n      <array>\n        <map>\n         <key>Channel</key>\n          <integer>0</integer>\n         <key>Message</key>\n          <string>test <!HIPPOEVAL[[\n            1 + 1\n          ]]></string>\n         <key>Type</key>\n          <integer>1</integer>\n        </map>\n      </array>\n    </map>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='UpdateAvatarAppearance', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>cof_version</key>\n    <integer></integer>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='ServerReleaseNotes', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='UserInfo', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='SendPostcard', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>msg</key>\n    <string></string>\n    <key>name</key>\n    <string></string>\n    <key>pos-global</key>\n    <array>\n      <real></real>\n      <real></real>\n      <real></real>\n    </array>\n    <key>subject</key>\n    <string></string>\n    <key>to</key>\n    <string></string>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='GetAdminExperiences', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='GetExperiences', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='AgentExperiences', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='ExperiencePreferences', method='GET', body=b'', query={'SOME_ID'}),
    CAPTemplate(cap_name='IsExperienceAdmin', method='GET', body=b'', query={'experience_id'}),
    CAPTemplate(cap_name='UpdateExperience', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>description</key>\n    <string></string>\n    <key>extended_metadata</key>\n    <string></string>\n    <key>group_id</key>\n    <uuid></uuid>\n    <key>slurl</key>\n    <string></string>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='AttachmentResources', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='DispatchRegionInfo', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>agent_limit</key>\n    <real></real>\n    <key>allow_damage</key>\n    <integer></integer>\n    <key>allow_land_resell</key>\n    <integer></integer>\n    <key>allow_parcel_changes</key>\n    <integer></integer>\n    <key>block_fly</key>\n    <integer></integer>\n    <key>block_fly_over</key>\n    <integer></integer>\n    <key>block_parcel_search</key>\n    <integer></integer>\n    <key>block_terraform</key>\n    <integer></integer>\n    <key>prim_bonus</key>\n    <real></real>\n    <key>restrict_pushobject</key>\n    <integer></integer>\n    <key>sim_access</key>\n    <string></string>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='AvatarPickerSearch', method='GET', body=b'', query={'names', 'page_size'}),
    CAPTemplate(cap_name='LandResources', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>parcel_id</key>\n    <uuid><!HIPPOREPL[[SELECTED_PARCEL_FULL]]></uuid>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='ResourceCostSelected', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>selected_roots</key>\n    <array>\n      <uuid><!HIPPOREPL[[SELECTED_FULL]]></uuid>\n    </array>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='GroupMemberData', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>group_id</key>\n    <uuid />\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='GroupExperiences', method='GET', body=b'', query={'SOME_ID'}),
    CAPTemplate(cap_name='CopyInventoryFromNotecard', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>callback-id</key>\n    <integer></integer>\n    <key>folder-id</key>\n    <uuid></uuid>\n    <key>item-id</key>\n    <uuid></uuid>\n    <key>notecard-id</key>\n    <uuid></uuid>\n    <key>object-id</key>\n    <uuid />\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='HomeLocation', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>HomeLocation</key>\n    <map>\n      <key>LocationId</key>\n      <integer>1</integer>\n      <key>LocationLookAt</key>\n      <map>\n        <key>X</key>\n        <real></real>\n        <key>Y</key>\n        <real></real>\n        <key>Z</key>\n        <real></real>\n      </map>\n      <key>LocationPos</key>\n      <map>\n        <key>X</key>\n        <real></real>\n        <key>Y</key>\n        <real></real>\n        <key>Z</key>\n        <real></real>\n      </map>\n    </map>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='CharacterProperties', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='ObjectNavMeshProperties', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='TerrainNavMeshProperties', method='GET', body=b'', query=set()),
    CAPTemplate(cap_name='ObjectNavMeshProperties', method='PUT', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key><!HIPPOREPL[[SELECTED_FULL]]></key>\n    <map>\n      <key>A</key>\n      <integer></integer>\n      <key>B</key>\n      <integer></integer>\n      <key>C</key>\n      <integer></integer>\n      <key>D</key>\n      <integer></integer>\n      <key>navmesh_category</key>\n      <integer></integer>\n      <key>phantom</key>\n      <boolean></boolean>\n    </map>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='RetrieveNavMeshSrc', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<undef />\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='MapLayer', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<undef />\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='GroupAPIv1', method='GET', body=b'', query={'group_id'}),
    CAPTemplate(cap_name='ChatSessionRequest', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>method</key>\n    <string></string>\n    <key>session-id</key>\n    <uuid></uuid>\n  </map>\n</llsd>\n', query=set()),
    CAPTemplate(cap_name='ViewerBenefits', method='GET', body=b'', query=set(), path=''),
    CAPTemplate(cap_name='SetDisplayName', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>display_name</key>\n    <array>\n      <string>OLD_DISPLAY_NAME</string>\n      <string>NEW_DISPLAY_NAME</string>\n    </array>\n  </map>\n</llsd>\n', query=set(), path=''),
    CAPTemplate(cap_name='ObjectMediaNavigate', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n    <key>current_url</key>\n    <string></string>\n    <key>object_id</key>\n    <uuid><!HIPPOREPL[[SELECTED_FULL]]></uuid>\n    <key>texture_index</key>\n    <integer></integer>\n  </map>\n</llsd>\n', query=set(), path=''),
    CAPTemplate(cap_name='AgentProfile', method='GET', body=b'', query=set(), path='/<SOME_ID>'),
    CAPTemplate(cap_name='InterestList', method='POST', body=b'<?xml version="1.0" ?>\n<llsd>\n<map>\n   <key>mode</key>\n    <string>360</string>\n  </map>\n</llsd>', query=set(), path='/'),
    CAPTemplate(cap_name='RegionObjects', method='GET', body=b'', query=set(), path=''),
]



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

from __future__ import annotations

import dataclasses
from typing import *


_T = TypeVar("_T")


class SettingDescriptor(Generic[_T]):
    __slots__ = ("name", "default")

    def __init__(self, default: Union[Callable[[], _T], _T]):
        self.default = default
        self.name: Optional[str] = None

    def __set_name__(self, owner: Settings, name: str):
        self.name = name

    def _make_default(self) -> _T:
        if callable(self.default):
            return self.default()
        return self.default

    def __get__(self, obj: Settings, owner: Optional[Type] = None) -> _T:
        val: Union[_T, dataclasses.MISSING] = obj.get_setting(self.name)
        if val is dataclasses.MISSING:
            val = self._make_default()
        return val

    def __set__(self, obj: Settings, value: _T) -> None:
        obj.set_setting(self.name, value)


class Settings:
    ENABLE_DEFERRED_PACKET_PARSING: bool = SettingDescriptor(True)

    def __init__(self):
        self._settings: Dict[str, Any] = {}

    def get_setting(self, name: str) -> Any:
        return self._settings.get(name, dataclasses.MISSING)

    def set_setting(self, name: str, val: Any):
        self._settings[name] = val



import itertools
import sys
from pathlib import Path

from hippolyzer.lib.base import llsd


def iter_viewer_data_dirs():
    if sys.platform.startswith("linux"):
        paths = (x for x in Path.home().iterdir() if x.name.startswith("."))
    elif sys.platform == "darwin":
        paths = (Path.home() / "Library" / "Application Support").iterdir()
    elif sys.platform in ("win32", "msys", "cygwin"):
        app_data = Path.home() / "AppData"
        # On Windows the cache directory is in Local, the settings are in Roaming. I think.
        paths = itertools.chain((app_data / "Local").iterdir(), (app_data / "Roaming").iterdir())
    else:
        raise Exception("Unknown OS, can't locate viewer config dirs!")

    for path in paths:
        if not path.is_dir():
            continue
        if not has_settings_file(path) and not has_cache_file(path):
            continue
        yield path


def has_cache_file(path: Path):
    try:
        return (path / "avatar_name_cache.xml").exists()
    except PermissionError:
        return False


def has_settings_file(path: Path):
    try:
        return (path / "user_settings" / "settings.xml").exists()
    except PermissionError:
        return False


def iter_viewer_config_dirs():
    for viewer_dir in iter_viewer_data_dirs():
        if has_settings_file(viewer_dir):
            yield viewer_dir


def iter_viewer_cache_dirs():
    for viewer_dir in iter_viewer_data_dirs():
        # Is this a settings dir
        if has_settings_file(viewer_dir):
            # Users can choose custom locations for the cache directory, we need to parse
            # their settings to see if they've done so.
            with open(viewer_dir / "user_settings" / "settings.xml", "rb") as fh:
                config: dict = llsd.parse_xml(fh.read())
            # TODO: is this the case on all platforms?
            cache_location = None
            cache_elem = config.get("CacheLocation")
            if cache_elem:
                cache_location = cache_elem.get("Value")
            if cache_location:
                cache_location = Path(cache_location)
                if has_cache_file(cache_location):
                    yield cache_location
        # Cache may be in the base dir on Windows
        if has_cache_file(viewer_dir):
            yield viewer_dir
        # but it might also be in a subfolder
        if has_cache_file(viewer_dir / "cache"):
            yield viewer_dir / "cache"



from __future__ import annotations

import logging
from typing import *

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.client.namecache import NameCache
from hippolyzer.lib.proxy.viewer_settings import iter_viewer_cache_dirs

if TYPE_CHECKING:
    from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow


class ProxyNameCache(NameCache):
    def create_subscriptions(
            self,
            message_handler: MessageHandler[Message, str],
            http_message_handler: Optional[MessageHandler[HippoHTTPFlow, str]] = None,
    ):
        super().create_subscriptions(message_handler)
        if http_message_handler is not None:
            http_message_handler.subscribe("GetDisplayNames", self._handle_get_display_names)

    def load_viewer_caches(self):
        for cache_dir in iter_viewer_cache_dirs():
            try:
                namecache_file = cache_dir / "avatar_name_cache.xml"
                if namecache_file.exists():
                    with open(namecache_file, "rb") as f:
                        namecache_bytes = f.read()
                    agents = llsd.parse_xml(namecache_bytes)["agents"]
                    # Can be `None` if the file was just created
                    if not agents:
                        continue
                    for agent_id, agent_data in agents.items():
                        # Don't set display name if they just have the default
                        display_name = None
                        if not agent_data["is_display_name_default"]:
                            display_name = agent_data["display_name"]
                        self.update(UUID(agent_id), {
                            "FirstName": agent_data["legacy_first_name"],
                            "LastName": agent_data["legacy_last_name"],
                            "DisplayName": display_name,
                        })
            except:
                logging.exception(f"Failed to load namecache from {cache_dir}")

    def _handle_get_display_names(self, flow: HippoHTTPFlow):
        if flow.response.status_code != 200:
            return
        self._process_display_names_response(llsd.parse_xml(flow.response.content))



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import enum


# represents how much of a message is taken up by packet ID things, such as
# the packet flags and the sequence number. After the ID, then comes the header
# NOTE: This will be moved into a messaging system eventually

class PacketLayout:
    PACKET_ID_LENGTH = 6
    PHL_FLAGS = 0
    PHL_PACKET_ID = 1  # length of 4
    PHL_OFFSET = 5
    PHL_NAME = 6
    # 1 byte flags, 4 bytes sequence, 1 byte offset + 1 byte message name (high)
    MINIMUM_VALID_PACKET_SIZE = PACKET_ID_LENGTH + 1


class EndianType:
    LITTLE = '<'
    BIG = '>'
    NETWORK = '!'
    NONE = ''


class MsgBlockType:
    MBT_SINGLE = 0
    MBT_MULTIPLE = 1
    MBT_VARIABLE = 2
    MBT_String_List = ['Single', 'Multiple', 'Variable']


class PacketFlags(enum.IntFlag):
    ZEROCODED = 0x80
    RELIABLE = 0x40
    RESENT = 0x20
    ACK = 0x10


# frequency for messages
# = '\xFF\xFF\xFF'
# = '\xFF\xFF'
# = '\xFF'
# = ''
class MsgFrequency:
    FIXED_FREQUENCY_MESSAGE = -1  # marking it
    LOW_FREQUENCY_MESSAGE = 4
    MEDIUM_FREQUENCY_MESSAGE = 2
    HIGH_FREQUENCY_MESSAGE = 1


class MsgTrust:
    LL_NOTRUST = 0
    LL_TRUSTED = 1


class MsgEncoding:
    LL_UNENCODED = 0
    LL_ZEROCODED = 1


class MsgDeprecation:
    LL_DEPRECATED = 0
    LL_UDPDEPRECATED = 1
    LL_UDPBLACKLISTED = 2
    LL_NOTDEPRECATED = 3


# message variable types
class MsgType(enum.Enum):
    MVT_FIXED = 0
    MVT_VARIABLE = enum.auto()
    MVT_U8 = enum.auto()
    MVT_U16 = enum.auto()
    MVT_U32 = enum.auto()
    MVT_U64 = enum.auto()
    MVT_S8 = enum.auto()
    MVT_S16 = enum.auto()
    MVT_S32 = enum.auto()
    MVT_S64 = enum.auto()
    MVT_F32 = enum.auto()
    MVT_F64 = enum.auto()
    MVT_LLVector3 = enum.auto()
    MVT_LLVector3d = enum.auto()
    MVT_LLVector4 = enum.auto()
    MVT_LLQuaternion = enum.auto()
    MVT_LLUUID = enum.auto()
    MVT_BOOL = enum.auto()
    MVT_IP_ADDR = enum.auto()
    MVT_IP_PORT = enum.auto()

    @property
    def is_int(self):
        return any(self == x for x in (
            self.MVT_U8,
            self.MVT_U16,
            self.MVT_U32,
            self.MVT_U64,
            self.MVT_S8,
            self.MVT_S16,
            self.MVT_S32,
            self.MVT_S64,
            self.MVT_IP_PORT,
        ))

    @property
    def is_float(self):
        return any(self == x for x in (
            self.MVT_F32,
            self.MVT_F64,
        ))

    @property
    def size(self):
        return TYPE_SIZES[self]


TYPE_SIZES = {
    MsgType.MVT_FIXED: -1,
    MsgType.MVT_VARIABLE: -1,
    MsgType.MVT_U8: 1,
    MsgType.MVT_U16: 2,
    MsgType.MVT_U32: 4,
    MsgType.MVT_U64: 8,
    MsgType.MVT_S8: 1,
    MsgType.MVT_S16: 2,
    MsgType.MVT_S32: 4,
    MsgType.MVT_S64: 8,
    MsgType.MVT_F32: 4,
    MsgType.MVT_F64: 8,
    MsgType.MVT_LLVector3: 12,
    MsgType.MVT_LLVector3d: 24,
    MsgType.MVT_LLVector4: 16,
    MsgType.MVT_LLQuaternion: 12,
    MsgType.MVT_LLUUID: 16,
    MsgType.MVT_BOOL: 1,
    MsgType.MVT_IP_ADDR: 4,
    MsgType.MVT_IP_PORT: 2,
}



"""
Body parts and linden clothing layers
"""

from __future__ import annotations

import dataclasses
import logging
from io import StringIO
from typing import *

from xml.etree.ElementTree import parse as parse_etree

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.helpers import get_resource_filename
from hippolyzer.lib.base.inventory import InventorySaleInfo, InventoryPermissions
from hippolyzer.lib.base.legacy_schema import SchemaBase, parse_schema_line, SchemaParsingError
from hippolyzer.lib.base.templates import WearableType

LOG = logging.getLogger(__name__)
_T = TypeVar("_T")

WEARABLE_VERSION = "LLWearable version 22"


@dataclasses.dataclass
class VisualParam:
    id: int
    name: str
    value_min: float
    value_max: float
    # These might be `None` if the param isn't meant to be directly edited
    edit_group: Optional[str]
    wearable: Optional[str]


class VisualParams(List[VisualParam]):
    def __init__(self, lad_path):
        super().__init__()
        with open(lad_path, "rb") as f:
            doc = parse_etree(f)
        for param in doc.findall(".//param"):
            self.append(VisualParam(
                id=int(param.attrib["id"]),
                name=param.attrib["name"],
                edit_group=param.get("edit_group"),
                wearable=param.get("wearable"),
                value_min=float(param.attrib["value_min"]),
                value_max=float(param.attrib["value_max"]),
            ))

    def by_name(self, name: str) -> VisualParam:
        return [x for x in self if x.name == name][0]

    def by_edit_group(self, edit_group: str) -> List[VisualParam]:
        return [x for x in self if x.edit_group == edit_group]

    def by_wearable(self, wearable: str) -> List[VisualParam]:
        return [x for x in self if x.wearable == wearable]

    def by_id(self, vparam_id: int) -> VisualParam:
        return [x for x in self if x.id == vparam_id][0]


VISUAL_PARAMS = VisualParams(get_resource_filename("lib/base/data/avatar_lad.xml"))


@dataclasses.dataclass
class Wearable(SchemaBase):
    name: str
    wearable_type: WearableType
    permissions: InventoryPermissions
    sale_info: InventorySaleInfo
    # VisualParam ID -> val
    parameters: Dict[int, float]
    # TextureEntry ID -> texture ID
    textures: Dict[int, UUID]

    @classmethod
    def _skip_to_next_populated_line(cls, reader: StringIO):
        old_pos = reader.tell()
        while peeked_data := reader.readline():
            # Read until we find a non-blank line
            if peeked_data.lstrip("\n"):
                break
            old_pos = reader.tell()
        # Reading an empty string means EOF
        if not peeked_data:
            raise SchemaParsingError("Premature EOF")
        reader.seek(old_pos)

    @classmethod
    def _read_and_parse_line(cls, reader: StringIO):
        cls._skip_to_next_populated_line(reader)
        return parse_schema_line(reader.readline())

    @classmethod
    def _read_expected_key(cls, reader: StringIO, expected_key: str) -> str:
        key, val = cls._read_and_parse_line(reader)
        if key != expected_key:
            raise ValueError(f"Expected {expected_key} not found, {(key, val)!r}")
        return val

    @classmethod
    def from_reader(cls, reader: StringIO) -> Wearable:
        cls._skip_to_next_populated_line(reader)
        version_str = reader.readline().rstrip()
        if version_str != WEARABLE_VERSION:
            raise ValueError(f"Bad wearable version {version_str!r}")
        cls._skip_to_next_populated_line(reader)
        name = reader.readline().rstrip()

        permissions = InventoryPermissions.from_reader(reader, read_header=True)
        sale_info = InventorySaleInfo.from_reader(reader, read_header=True)

        wearable_type = WearableType(int(cls._read_expected_key(reader, "type")))
        num_params = int(cls._read_expected_key(reader, "parameters"))
        params = {}
        for _ in range(num_params):
            param_id, param_val = cls._read_and_parse_line(reader)
            if param_val == ".":
                param_val = "0.0"
            params[int(param_id)] = float(param_val)

        num_textures = int(cls._read_expected_key(reader, "textures"))
        textures = {}
        for _ in range(num_textures):
            te_id, texture_id = cls._read_and_parse_line(reader)
            textures[int(te_id)] = UUID(texture_id)
        return Wearable(
            name=name,
            wearable_type=wearable_type,
            permissions=permissions,
            sale_info=sale_info,
            parameters=params,
            textures=textures
        )

    def to_writer(self, writer: StringIO):
        writer.write(f"{WEARABLE_VERSION}\n")
        writer.write(f"{self.name}\n\n")
        self.permissions.to_writer(writer)
        self.sale_info.to_writer(writer)
        writer.write(f"type {int(self.wearable_type)}\n")
        writer.write(f"parameters {len(self.parameters)}\n")
        for param_id, param_val in self.parameters.items():
            writer.write(f"{param_id} {param_val}\n")
        writer.write(f"textures {len(self.textures)}\n")
        for te_id, texture_id in self.textures.items():
            writer.write(f"{te_id} {texture_id}\n")



"""
Managers for inbound and outbound xfer as well as the AssetUploadRequest flow
"""
from __future__ import annotations

import asyncio
import enum
import random
from typing import *

from hippolyzer.lib.base.datatypes import UUID, RawBytes
from hippolyzer.lib.base.message.data_packer import TemplateDataPacker
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.message.msgtypes import MsgType, PacketFlags
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.base.message.circuit import ConnectionHolder
from hippolyzer.lib.base.templates import XferPacket, XferFilePath, AssetType, XferError

_XFER_MESSAGES = {"AbortXfer", "ConfirmXferPacket", "RequestXfer", "SendXferPacket"}


MAX_CHUNK_SIZE = 1150
ACK_AHEAD_MAX = 10


class Xfer:
    def __init__(
            self,
            xfer_id: Optional[int] = None,
            direction: Direction = Direction.OUT,
            data: Optional[bytes] = None,
            turbo: bool = False,
    ):
        self.xfer_id: Optional[int] = xfer_id
        self.chunks: Dict[int, bytes] = {}
        self.expected_size: Optional[int] = None
        self.size_known = asyncio.Future()
        self.error_code: Union[int, XferError] = 0
        self.next_ackable = 0
        self.turbo = turbo
        self.direction: Direction = direction
        self.expected_chunks: Optional[int] = None
        self._future: asyncio.Future[Xfer] = asyncio.Future()

        if data is not None:
            # Prepend the expected length field to the first chunk
            if not isinstance(data, RawBytes):
                data = TemplateDataPacker.pack(len(data), MsgType.MVT_S32) + data
            chunk_num = 0
            while data:
                self.chunks[chunk_num] = data[:MAX_CHUNK_SIZE]
                data = data[MAX_CHUNK_SIZE:]
                chunk_num += 1

    def reassemble_chunks(self) -> bytes:
        assembled = bytearray()
        for _, data in sorted(self.chunks.items()):
            assembled.extend(data)
        return assembled

    def mark_done(self):
        self._future.set_result(self)

    def done(self) -> bool:
        return self._future.done()

    def cancelled(self) -> bool:
        return self._future.cancelled()

    def is_our_message(self, message):
        return message["XferID"]["ID"] == self.xfer_id

    def cancel(self) -> bool:
        if not self.size_known.done():
            self.size_known.cancel()
        return self._future.cancel()

    def set_exception(self, exc: Union[type, BaseException]) -> None:
        if not self.size_known.done():
            self.size_known.set_exception(exc)
        return self._future.set_exception(exc)

    def __await__(self) -> Generator[Any, None, Xfer]:
        return self._future.__await__()


class UploadStrategy(enum.IntEnum):
    XFER = enum.auto()
    ASSET_UPLOAD_REQUEST = enum.auto()


class XferManager:
    def __init__(
            self,
            connection_holder: ConnectionHolder,
            secure_session_id: Optional[UUID] = None,
    ):
        self._connection_holder = connection_holder
        self._secure_session_id = secure_session_id

    def request(
            self, xfer_id: Optional[int] = None,
            file_name: Union[bytes, str, None] = None,
            file_path: Optional[Union[XferFilePath, int]] = None,
            vfile_id: Optional[UUID] = None,
            vfile_type: Optional[Union[AssetType, int]] = None,
            use_big_packets: bool = False,
            delete_on_completion: bool = True,
            turbo: bool = False,
            direction: Direction = Direction.OUT,
    ) -> Xfer:
        xfer_id = xfer_id if xfer_id is not None else random.getrandbits(64)
        self._connection_holder.circuit.send(Message(
            'RequestXfer',
            Block(
                'XferID',
                ID=xfer_id,
                Filename=file_name or b'',
                FilePath=file_path or XferFilePath.NONE,
                DeleteOnCompletion=delete_on_completion,
                UseBigPackets=use_big_packets,
                VFileID=vfile_id or UUID(),
                VFileType=vfile_type or AssetType.NONE,
            ),
            direction=direction,
        ))
        xfer = Xfer(xfer_id, direction=direction, turbo=turbo)
        asyncio.create_task(self._pump_xfer_replies(xfer))
        return xfer

    async def _pump_xfer_replies(self, xfer: Xfer):
        with self._connection_holder.message_handler.subscribe_async(
                _XFER_MESSAGES,
                predicate=xfer.is_our_message,
        ) as get_msg:
            while not xfer.done():
                try:
                    msg: Message = await asyncio.wait_for(get_msg(), 5.0)
                except asyncio.exceptions.TimeoutError as e:
                    xfer.set_exception(e)
                    return

                if xfer.cancelled():
                    # AbortXfer doesn't seem to work on in-progress Xfers.
                    # Just let any new packets drop on the floor.
                    return

                if msg.name == "SendXferPacket":
                    self._handle_send_xfer_packet(msg, xfer)
                elif msg.name == "AbortXfer":
                    xfer.error_code = msg["XferID"][0].deserialize_var("Result")
                    xfer.set_exception(
                        ConnectionAbortedError(f"Xfer failed with {xfer.error_code!r}")
                    )

    def _handle_send_xfer_packet(self, msg: Message, xfer: Xfer):
        # Received a SendXfer for an Xfer we sent ourselves
        packet_id: XferPacket = msg["XferID"][0].deserialize_var("Packet")
        packet_data = msg["DataPacket"]["Data"]
        # First 4 bytes are expected total data length
        if packet_id.PacketID == 0:
            # Yes, S32. Only used as a hint so buffers can be pre-allocated,
            # EOF bit determines when the data actually ends.
            xfer.expected_size = TemplateDataPacker.unpack(packet_data[:4], MsgType.MVT_S32)
            # Don't re-set if we get a resend of packet 0
            if not xfer.size_known.done():
                xfer.size_known.set_result(xfer.expected_size)
            packet_data = packet_data[4:]

        to_ack = (packet_id.PacketID,)
        if xfer.turbo:
            # ACK the next few packets we expect to be sent, if we haven't already
            ack_max = packet_id.PacketID + ACK_AHEAD_MAX
            to_ack = range(xfer.next_ackable, ack_max)
            xfer.next_ackable = ack_max
        for ack_id in to_ack:
            self._connection_holder.circuit.send_reliable(Message(
                "ConfirmXferPacket",
                Block("XferID", ID=xfer.xfer_id, Packet=ack_id),
                direction=xfer.direction,
                flags=PacketFlags.RELIABLE,
            ))

        xfer.chunks[packet_id.PacketID] = packet_data
        # We may be waiting on other packets so we can't end immediately.
        if packet_id.IsEOF:
            xfer.expected_chunks = packet_id.PacketID + 1
        if not xfer.done() and len(xfer.chunks) == xfer.expected_chunks:
            xfer.mark_done()

    def upload_asset(
            self,
            asset_type: AssetType,
            data: Union[bytes, str],
            store_local: bool = False,
            temp_file: bool = False,
            transaction_id: Optional[UUID] = None,
            upload_strategy: Optional[UploadStrategy] = None,
    ) -> asyncio.Future[UUID]:
        """Upload an asset through the Xfer upload path"""
        if not transaction_id:
            transaction_id = UUID.random()
        if isinstance(data, str):
            data = data.encode("utf8")

        # Small amounts of data can be sent inline, decide based on size
        if upload_strategy is None:
            if len(data) >= MAX_CHUNK_SIZE:
                upload_strategy = UploadStrategy.XFER
            else:
                upload_strategy = UploadStrategy.ASSET_UPLOAD_REQUEST

        xfer = None
        inline_data = b''
        if upload_strategy == UploadStrategy.XFER:
            xfer = Xfer(data=data)
        else:
            inline_data = data

        self._connection_holder.circuit.send(Message(
            "AssetUploadRequest",
            Block(
                "AssetBlock",
                TransactionID=transaction_id,
                Type=asset_type,
                Tempfile=temp_file,
                StoreLocal=store_local,
                AssetData=inline_data,
            ),
            flags=PacketFlags.RELIABLE
        ))
        fut = asyncio.Future()
        asyncio.create_task(self._pump_asset_upload(xfer, transaction_id, fut))
        return fut

    async def _pump_asset_upload(self, xfer: Optional[Xfer], transaction_id: UUID, fut: asyncio.Future):
        message_handler = self._connection_holder.message_handler
        # We'll receive an Xfer request for the asset we're uploading.
        # asset ID is determined by hashing secure session ID with chosen transaction ID.
        asset_id: UUID = UUID.combine(transaction_id, self._secure_session_id)
        try:
            # Only need to do this if we're using the xfer upload strategy, otherwise all the
            # data was already sent in the AssetUploadRequest and we don't expect a RequestXfer.
            def request_predicate(request_msg: Message):
                return request_msg["XferID"]["VFileID"] == asset_id
            if xfer is not None:
                await self.serve_inbound_xfer_request(xfer, request_predicate)

            def complete_predicate(complete_msg: Message):
                return complete_msg["AssetBlock"]["UUID"] == asset_id
            msg = await message_handler.wait_for(('AssetUploadComplete',), predicate=complete_predicate)
            if msg["AssetBlock"]["Success"] == 1:
                fut.set_result(asset_id)
            else:
                fut.set_exception(RuntimeError(f"Xfer for transaction {transaction_id} failed"))

        except asyncio.TimeoutError as e:
            fut.set_exception(e)

    async def serve_inbound_xfer_request(
            self,
            xfer: Xfer,
            request_predicate: Callable[[Message], bool],
            wait_for_confirm: bool = True
    ):
        message_handler = self._connection_holder.message_handler
        request_msg = await message_handler.wait_for(
            ('RequestXfer',), predicate=request_predicate, timeout=5.0)
        xfer.xfer_id = request_msg["XferID"]["ID"]

        packet_id = 0
        # TODO: No resend yet. If it's lost, it's lost.
        while xfer.chunks:
            chunk = xfer.chunks.pop(packet_id)
            # EOF if there are no chunks left
            packet_val = XferPacket(PacketID=packet_id, IsEOF=not bool(xfer.chunks))
            self._connection_holder.circuit.send(Message(
                "SendXferPacket",
                Block("XferID", ID=xfer.xfer_id, Packet_=packet_val),
                Block("DataPacket", Data=chunk),
                # Send this towards the sender of the RequestXfer
                direction=~request_msg.direction,
                flags=PacketFlags.RELIABLE,
            ))
            # Don't care about the value, just want to know it was confirmed.
            if wait_for_confirm:
                await message_handler.wait_for(
                    ("ConfirmXferPacket",), predicate=xfer.is_our_message, timeout=5.0)
            packet_id += 1



from __future__ import annotations

import copy
import multiprocessing
import weakref
from typing import *
from typing import Optional

import mitmproxy.http
from mitmproxy.http import HTTPFlow

from hippolyzer.lib.proxy.caps import CapData

if TYPE_CHECKING:
    from hippolyzer.lib.proxy.sessions import SessionManager


class HippoHTTPFlow:
    """
    Wrapper for Hippolyzer-side mitmproxy flows

    Hides the nastiness of writing to flow.metadata so we can pass
    state back and forth between the two proxies
    """
    __slots__ = ("flow", "callback_queue", "resumed", "taken")

    def __init__(self, flow: HTTPFlow, callback_queue: Optional[multiprocessing.Queue] = None):
        self.flow: HTTPFlow = flow
        self.resumed = False
        self.taken = False
        self.callback_queue = weakref.ref(callback_queue) if callback_queue else None
        meta = self.flow.metadata
        meta.setdefault("can_stream", True)
        meta.setdefault("response_injected", False)
        meta.setdefault("request_injected", False)
        meta.setdefault("cap_data", CapData())
        meta.setdefault("from_browser", False)

    @property
    def request(self) -> mitmproxy.http.Request:
        return self.flow.request

    @property
    def response(self) -> Optional[mitmproxy.http.Response]:
        return self.flow.response

    @property
    def id(self) -> str:
        return self.flow.id

    @response.setter
    def response(self, val: Optional[mitmproxy.http.Response]):
        self.flow.metadata["response_injected"] = True
        self.flow.response = val

    @property
    def response_injected(self) -> bool:
        return self.flow.metadata["response_injected"]

    @property
    def request_injected(self) -> bool:
        # Populated by mitmproxy side based on X-Hippo-Injected header
        return self.flow.metadata["request_injected"]

    @property
    def metadata(self) -> Dict[str, Any]:
        return self.flow.metadata

    @property
    def cap_data(self) -> Optional[CapData]:
        return self.metadata["cap_data"]

    @cap_data.setter
    def cap_data(self, val: Optional[CapData]):
        self.metadata["cap_data"] = val

    @property
    def can_stream(self) -> bool:
        return self.metadata["can_stream"]

    @can_stream.setter
    def can_stream(self, val: bool):
        # can != will, only applies to asset server reqs
        self.metadata["can_stream"] = val

    @property
    def from_browser(self) -> bool:
        return self.metadata["from_browser"]

    @property
    def name(self) -> Optional[str]:
        if self.cap_data:
            return self.cap_data.cap_name
        return None

    def take(self) -> HippoHTTPFlow:
        """Don't automatically pass this flow back to mitmproxy"""
        # TODO: Having to explicitly take / release Flows to use them in an async
        #  context is kind of janky. The HTTP callback handling code should probably
        #  be made totally async, including the addon hooks. Would coroutine per-callback
        #  be expensive?
        assert not self.taken and not self.resumed
        self.taken = True
        return self

    def resume(self):
        """Release the HTTP flow back to the normal processing flow"""
        assert self.callback_queue
        assert not self.resumed
        self.taken = False
        self.resumed = True
        self.callback_queue().put(("callback", self.flow.id, self.get_state()))

    def preempt(self):
        # Must be some flow that we previously resumed, we're racing
        # the result from the server end.
        assert not self.taken and self.resumed
        self.callback_queue().put(("preempt", self.flow.id, self.get_state()))

    @property
    def is_replay(self) -> bool:
        return bool(self.flow.is_replay)

    def get_state(self) -> Dict:
        flow = self.flow
        # Not serializable, so we have to pop it off to send across the wire.
        cap_data: Optional[CapData] = flow.metadata.pop("cap_data", None)
        if cap_data is not None:
            flow.metadata["cap_data_ser"] = cap_data.serialize()
        else:
            flow.metadata["cap_data_ser"] = None
        state = self.flow.get_state()
        # Shove it back on
        flow.metadata["cap_data"] = cap_data
        return state

    @classmethod
    def from_state(cls, flow_state: Dict, session_manager: Optional[SessionManager]) -> HippoHTTPFlow:
        flow: Optional[HTTPFlow] = HTTPFlow.from_state(flow_state)
        assert flow is not None
        cap_data_ser = flow.metadata.get("cap_data_ser")
        callback_queue = None
        if session_manager:
            callback_queue = session_manager.flow_context.to_proxy_queue
        if cap_data_ser is not None:
            flow.metadata["cap_data"] = CapData.deserialize(cap_data_ser, session_manager)
        else:
            flow.metadata["cap_data"] = None
        return cls(flow, callback_queue)

    def copy(self) -> HippoHTTPFlow:
        # HACK: flow.copy() expects the flow to be fully JSON serializable, but
        # our cap data won't be due to the session objects. Deal with that manually.
        flow = self.flow
        cap_data = flow.metadata.pop("cap_data")
        new_flow = self.__class__(self.flow.copy())
        flow.metadata["cap_data"] = cap_data
        new_flow.metadata["cap_data"] = copy.copy(cap_data)
        return new_flow
